from flask_restx import Namespace, Resource, fields, reqparse
from flask import request, make_response
from .... import PyAutomation
from ....extensions.api import api
from ....extensions import _api as Api
import json

ns = Namespace('Settings', description='Application Configuration Settings')
app = PyAutomation()

settings_model = api.model("settings_model", {
    'logger_period': fields.Float(required=False, min=1.0, description='Logger worker period in seconds (>= 1.0)'),
    'log_max_bytes': fields.Integer(required=False, min=1024, description='Max bytes for log file rotation (>= 1024)'),
    'log_backup_count': fields.Integer(required=False, min=1, description='Number of backup log files to keep (>= 1)'),
    'log_level': fields.Integer(required=False, min=0, max=50, description='Logging level (0=NOTSET, 10=DEBUG, 20=INFO, 30=WARNING, 40=ERROR, 50=CRITICAL)')
})

@ns.route('/update')
class SettingsUpdateResource(Resource):
    
    @api.doc(security='apikey', description="Updates various application settings.")
    @api.response(200, "Settings updated successfully")
    @api.response(400, "Invalid parameter values")
    @Api.token_required(auth=True)
    @ns.expect(settings_model)
    def put(self):
        """
        Update settings.

        Updates application configuration including logger period, log rotation settings, and logging level.
        """
        data = api.payload
        
        # 1. Update Logger Worker Period
        if 'logger_period' in data:
            logger_period = data['logger_period']
            if logger_period < 1.0:
                return "Logger period must be >= 1.0", 400
            app.update_logger_period(logger_period)

        # 2. Update Log Rotation Config
        if 'log_max_bytes' in data and 'log_backup_count' in data:
            max_bytes = data['log_max_bytes']
            backup_count = data['log_backup_count']
            
            if max_bytes < 1024:
                 return "log_max_bytes must be >= 1024", 400
            if backup_count < 1:
                 return "log_backup_count must be >= 1", 400
                 
            app.update_log_config(max_bytes, backup_count)
        
        elif 'log_max_bytes' in data or 'log_backup_count' in data:
             return "Both log_max_bytes and log_backup_count must be provided together", 400

        # 3. Update Log Level
        if 'log_level' in data:
            log_level = data['log_level']
            # Basic validation for standard levels
            if log_level not in [0, 10, 20, 30, 40, 50]:
                return "Invalid log_level. Use standard Python logging levels (10, 20, 30, 40, 50)", 400
            
            app.update_log_level(log_level)

        return "Settings updated", 200

@ns.route('/logger_period')
class LoggerPeriodResource(Resource):
    
    @api.doc(security='apikey', description="Updates only the logger worker period.")
    @api.response(200, "Logger period updated successfully")
    @api.response(400, "Invalid logger period")
    @Api.token_required(auth=True)
    @ns.expect(settings_model)
    def put(self):
        """
        Update logger period.

        Updates the execution period of the logger worker.
        """
        logger_period = api.payload.get('logger_period')
        
        if logger_period < 1.0:
            return "Logger period must be >= 1.0", 400
            
        app.update_logger_period(logger_period)
        
        return "Logger period updated", 200

@ns.route('/export_config')
class ExportConfigResource(Resource):
    
    @api.doc(security='apikey', description="Exports all configuration data to a JSON file. Excludes historical data (TagValue, Events, Logs, AlarmSummary).")
    @api.response(200, "Configuration exported successfully")
    @api.response(400, "Export failed")
    @api.response(401, "Unauthorized")
    # @Api.token_required(auth=True)
    def get(self):
        """
        Export configuration.

        Exports all configuration tables (Manufacturer, Segment, Variables, Units, DataTypes,
        Tags, AlarmTypes, AlarmStates, Alarms, Roles, Users, OPCUA, AccessType, OPCUAServer,
        Machines, TagsMachines) to a JSON file. Historical data is excluded.
        """
        try:
            config_data = app.export_configuration()
            
            if "error" in config_data:
                return {'message': config_data["error"]}, 400
            
            # Create JSON string in memory
            json_str = json.dumps(config_data, indent=2, default=str)
            json_bytes = json_str.encode('utf-8')
            
            # Create response with explicit headers to force download
            # Using application/octet-stream helps force download in Swagger UI
            response = make_response(json_bytes)
            response.headers['Content-Type'] = 'application/octet-stream'
            response.headers['Content-Disposition'] = 'attachment; filename="configuration_export.json"'
            response.headers['Content-Length'] = str(len(json_bytes))
            response.headers['X-Content-Type-Options'] = 'nosniff'
            
            return response
        except Exception as e:
            return {'message': f'Export failed: {str(e)}'}, 400

import_config_parser = reqparse.RequestParser(bundle_errors=True)
import_config_parser.add_argument('file', type=reqparse.FileStorage, location='files', required=True, help='JSON configuration file to import')

@ns.route('/import_config')
class ImportConfigResource(Resource):
    
    @Api.validate_reqparser(reqparser=import_config_parser)
    @api.doc(security='apikey', description="Imports configuration data from a JSON file. Restores all configuration tables while preserving historical data.")
    @api.response(200, "Configuration imported successfully")
    @api.response(400, "Import failed")
    @api.response(401, "Unauthorized")
    @ns.expect(import_config_parser)
    # @Api.token_required(auth=True)
    def post(self):
        """
        Import configuration.

        Imports configuration data from a JSON file. The file should be in the format
        exported by the export_config endpoint. Historical data (TagValue, Events, Logs,
        AlarmSummary) is not affected by the import.
        """
        try:
            if 'file' not in request.files:
                return {'message': 'No file provided'}, 400
            
            file = request.files['file']
            
            if file.filename == '':
                return {'message': 'No file selected'}, 400
            
            if not file.filename.endswith('.json'):
                return {'message': 'File must be a JSON file'}, 400
            
            # Read and parse JSON
            file_content = file.read()
            try:
                config_data = json.loads(file_content.decode('utf-8'))
            except json.JSONDecodeError as e:
                return {'message': f'Invalid JSON file: {str(e)}'}, 400
            
            # Import configuration
            result = app.import_configuration(config_data)
            
            if "error" in result:
                return {'message': result["error"], 'details': result.get("results", {})}, 400
            
            return {
                'message': result.get("message", "Configuration imported successfully"),
                'summary': result.get("summary", {}),
                'results': result.get("results", {})
            }, 200
            
        except Exception as e:
            return {'message': f'Import failed: {str(e)}'}, 400



from ....extensions.api import api


def init_app():

    from .events import ns as ns_events
    from .logs import ns as ns_logs

    api.add_namespace(ns_events, path="/events")
    api.add_namespace(ns_logs, path="/logs")


# -*- coding: utf-8 -*-
"""workers/worker.py

This module implements the base class for all worker threads in the system.
"""
from threading import Thread
from threading import Event as ThreadEvent


class BaseWorker(Thread):
    r"""
    Base class for all worker threads.

    It extends `threading.Thread` and provides a standardized mechanism for stopping the thread
    using a `threading.Event`.
    """

    def  __init__(self):

        super(BaseWorker, self).__init__()

        self.stop_event = ThreadEvent()

    def get_stop_event(self):
        r"""
        Returns the stop event object.

        **Returns:**

        * **threading.Event**: The event used to signal the thread to stop.
        """
        return self.stop_event

    def stop(self):
        r"""
        Signals the worker thread to stop execution.
        """
        self.stop_event.set()

    def __getstate__(self):

        state = self.__dict__.copy()
        del state['stop_event']
        return state

    def __setstate__(self, state):
        
        self.__dict__.update(state)
        self.stop_event = ThreadEvent()



from ..utils.units import *

class Pressure(EngUnit):
    """Creates a pressure object that can store a pressure value and 
    convert between units of pressure.

    :param value: [int|float] Engineering value\n
    :param unit: [str] Engineering unit\n
    :return: [Pressure Object]\n

    ```python
    >>> from automation.variables.pressure import Pressure
    >>> pressure = Pressure(value=1.0, unit="atm")
    >>> pressure.value
    1.0
    >>> pressure.unit
    'atm'
    >>> pressure.convert(to_unit="psi")
    14.695979321588414
    >>> Pressure.convert_value(value=1.0, from_unit="atm", to_unit="bar")
    1.0132502738308866
    >>> Pressure.convert_values(values=[1.0, 10.0], from_unit="atm", to_unit="bar")
    [1.0132502738308866, 10.132502738308865]
    >>> pressure.change_unit(unit="psi")
    14.695979321588414
    >>> pressure.unit
    'psi'
    >>> pressure.get_value()
    [14.695979321588414, 'psi']
    >>> print(pressure)
    14.695979321588414 psi
    >>> pressure2 = Pressure(value=3.0, unit="bar")
    >>> pressure_result = pressure + pressure2
    >>> print(pressure_result)
    58.20737932158842 psi
    >>> pressure_result = pressure * pressure2
    >>> print(pressure_result)
    639.4426346533621 psi
    >>> pressure_result = pressure / pressure2
    >>> print(pressure_result)
    0.3377500912769622 psi
    >>> pressure_result = pressure ** pressure2
    >>> print(pressure_result)
    6.115644251483115e+50 psi

    ```

    """
    
    class Units(UnitSerializer):
        bar = 'bar' 
        mbar = 'mbar' 
        ubar = 'ubar' 
        Pa = 'Pa' 
        hPa = 'hPa' 
        kPa = 'kPa' 
        MPa = 'MPa' 
        kgcm2 = 'kgcm2' 
        atm = 'atm' 
        mmHg = 'mmHg' 
        mmH2O = 'mmH2O' 
        mH2O = 'mH2O' 
        psi = 'psi' 
        ftH2O = 'ftH2O' 
        inH2O = 'inH2O' 
        inHg = 'inHg' 
    
    conversions = {
        'bar' : 1.0,
        'mbar' : 1000.0,
        'ubar' : 1000000.0,
        'Pa' : 100000.0,
        'hPa' : 1000.0,
        'kPa' : 100.0,
        'MPa' : 0.1,
        'kgcm2' : 1.01972,
        'atm' : 0.986923,
        'mmHg' : 750.062,
        'mmH2O' : 10197.162129779,
        'mH2O' : 10.197162129779,
        'psi' : 14.5038,
        'ftH2O' : 33.455256555148,
        'inH2O' : 401.865,
        'inHg' : 29.53
    }

    def __init__(self, value, unit):

        if unit not in Pressure.Units.list():

            raise UnitError(f"{unit} value is not allowed for {self.__class__.__name__} object - you can use: {Pressure.Units.list()}")
        
        super(Pressure, self).__init__(value=value, unit=unit)


# -*- coding: utf-8 -*-
"""automation/logger/opcua_server.py

This module implements the OPC UA Server Logger, responsible for persisting
OPC UA Server node configurations and access rights.
"""
from ..dbmodels import OPCUAServer
from .core import BaseEngine, BaseLogger
from ..utils.decorators import db_rollback


class OPCUAServerLogger(BaseLogger):
    r"""
    Logger class specialized for OPC UA Server persistence.
    """

    def __init__(self):

        super(OPCUAServerLogger, self).__init__()

    @db_rollback
    def create(
            self,
            name:str,
            namespace:str,
            access_type:str="Read"
            ):
        r"""
        Creates a new OPC UA Node configuration in the database.

        **Parameters:**

        * **name** (str): Node name.
        * **namespace** (str): Node ID/Namespace.
        * **access_type** (str): Access rights ("Read", "Write", "ReadWrite").
        """
        if not self.check_connectivity():

            return None
       
        OPCUAServer.create(
            name=name,
            namespace=namespace,
            access_type=access_type
        )

    @db_rollback
    def put(
        self,
        namespace:str,
        access_type:str
        ):
        r"""
        Updates the access type for a specific OPC UA Node.

        **Parameters:**

        * **namespace** (str): Node ID/Namespace.
        * **access_type** (str): New access type.

        **Returns:**

        * **OPCUAServer**: The updated model instance.
        """
        if not self.check_connectivity():
            
            return None    
        
        if access_type:
            
            OPCUAServer.update_access_type(namespace=namespace, access_type=access_type)

            obj = OPCUAServer.read_by_namespace(namespace=namespace)

            return obj
    
    @db_rollback
    def read_all(self):
        r"""
        Retrieves all OPC UA Server nodes.
        """
        if not self.check_connectivity():

            return list()
        
        return OPCUAServer.read_all()
    
    @db_rollback
    def read_by_namespace(self, namespace:str):
        r"""
        Retrieves an OPC UA Server node by its namespace.
        """
        if not self.check_connectivity():

            return list()
        
        return OPCUAServer.read_by_namespace(namespace=namespace)
     

class OPCUAServerLoggerEngine(BaseEngine):
    r"""
    Thread-safe Engine for the OPCUAServerLogger.
    """

    def __init__(self):

        super(OPCUAServerLoggerEngine, self).__init__()
        self.logger = OPCUAServerLogger()

    def create(
        self,
        name:str,
        namespace:str,
        access_type:str="Read"
        ):
        r"""
        Thread-safe node creation.
        """
        _query = dict()
        _query["action"] = "create"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        _query["parameters"]["namespace"] = namespace
        _query["parameters"]["access_type"] = access_type
        
        return self.query(_query)
    
    def put(
        self,
        namespace:str,
        access_type:str
        ):
        r"""
        Thread-safe node update.
        """
        _query = dict()
        _query["action"] = "put"
        _query["parameters"] = dict()
        _query["parameters"]["namespace"] = namespace
        _query["parameters"]["access_type"] = access_type

        return self.query(_query)
    
    def read_by_namespace(
        self,
        namespace:str
        ):
        r"""
        Thread-safe read by namespace.
        """
        _query = dict()
        _query["action"] = "read_by_namespace"
        _query["parameters"] = dict()
        _query["parameters"]["namespace"] = namespace

        return self.query(_query)

    def read_all(self):
        r"""
        Thread-safe read all.
        """
        _query = dict()
        _query["action"] = "read_all"
        _query["parameters"] = dict()
        return self.query(_query)



import dash
import plotly.graph_objects as go

fig = go.Figure()

class TrendsComponents:

    @classmethod
    def tags(cls)->dash.dcc.Dropdown:
        r"""
        Documentation here
        """
        return dash.dcc.Dropdown(
            options = [],
            multi=True,
            id="trends_tags_dropdown",
            persistence=True
        )

    @classmethod
    def last_values(cls)->dash.dcc.Dropdown:
        r"""
        Documentation here
        """
        return dash.dcc.Dropdown(
            options=[
                {'label': 'last 10 seconds', 'value': 10},
                {'label': 'last 30 seconds', 'value': 30},
                {'label': 'last minute', 'value': 60},
                {'label': 'last 2 min.', 'value': 2 * 60},
                {'label': 'last 5 min.', 'value': 5 * 60},
                {'label': 'last 10 min.', 'value': 10 * 60}
            ],
            value=10,
            clearable=False,
            id="trends_last_values_dropdown"
        )

    @classmethod
    def current_value_table(cls)->dash.dash_table.DataTable:
        r"""
        Documentation here
        """
        return dash.dash_table.DataTable(
            data=[],
            columns=[ 
                {'name': 'tag', 'id': 'tag'}, 
                {'name': 'value', 'id': 'value'}
            ],
            id="trends_cvt_datatable",
            selected_columns=[],
            page_action="native",
            page_current= 0,
            page_size= 10,
            persistence=True,
            persisted_props=['data']
        )

    @classmethod
    def plot(cls)->dash.dcc.Graph:
        r"""
        Documentation here
        """
        return dash.dcc.Graph(
            figure=fig,
            id="trends_figure")


import os, pytz
from flask import Flask
from .core import PyAutomation
from .state_machine import OPCUAServer

app = Flask(__name__, instance_relative_config=False)

MANUFACTURER = os.environ.get('AUTOMATION_MANUFACTURER')
SEGMENT = os.environ.get('AUTOMATION_SEGMENT')
_TIMEZONE = os.environ.get('AUTOMATION_TIMEZONE') or "America/Caracas"
TIMEZONE = pytz.timezone(_TIMEZONE)
CERT_FILE = os.path.join(".", "ssl", os.environ.get('AUTOMATION_CERT_FILE') or "")
KEY_FILE = os.path.join(".", "ssl", os.environ.get('AUTOMATION_KEY_FILE') or "")
if not os.path.isfile(CERT_FILE):
    CERT_FILE = None

if not os.path.isfile(KEY_FILE):
    KEY_FILE = None
AUTOMATION_OPCUA_SERVER_PORT = os.environ.get('AUTOMATION_OPCUA_SERVER_PORT') or "53530"
AUTOMATION_LOGGER_PERIOD = os.environ.get('AUTOMATION_LOGGER_PERIOD') or 10.0
AUTOMATION_APP_SECRET_KEY = os.environ.get('AUTOMATION_APP_SECRET_KEY') or "073821603fcc483f9afee3f1500782a4"
AUTOMATION_SUPERUSER_PASSWORD = os.environ.get('AUTOMATION_SUPERUSER_PASSWORD') or "super_ultra_secret_password"


class CreateApp():
    """Initialize the core application."""

    def __call__(self):
        """
        Documentation here
        """
        app.client = None
        self.application = app
        
        with app.app_context():

            from . import extensions
            extensions.init_app(app)

            from . import modules
            modules.init_app(app)
            
            return app
        
__application = CreateApp()
server = __application()    
server.config['AUTOMATION_APP_SECRET_KEY'] = AUTOMATION_APP_SECRET_KEY
server.config['AUTOMATION_SUPERUSER_PASSWORD'] = AUTOMATION_SUPERUSER_PASSWORD
server.config['BUNDLE_ERRORS'] = True
opcua_server = OPCUAServer()



from .temperature import Temperature
from .length import Length
from .current import Current
from .eng_time import Time
from .pressure import Pressure
from .mass import Mass
from .force import Force
from .power import Power
from .volumetric_flow import VolumetricFlow
from .mass_flow import MassFlow
from .density import Density
from .percentage import Percentage
from .adimentional import Adimentional
from .volume import Volume


temperature_base = Temperature(value=1, unit='K')
length_base = Length(value=1, unit="m")
current_base = Current(value=1, unit="A")
time_base = Time(value=1, unit="s")
pressure_base = Pressure(value=1, unit="bar")
mass_base = Mass(value=1, unit="kg")
force_base = Force(value=1, unit="J/m")
power_base = Power(value=1, unit="kW")
volumetric_flow_base = VolumetricFlow(value=1, unit="bbl/sec")
mass_flow_base = MassFlow(value=1.0, unit="kg/day")
density_base = Density(value=1.0, unit="kg/bbl")
percentage_base = Percentage(value=0.0, unit="%")
adimentional_base = Adimentional(value=0.0, unit="adim")
volume_base = Volume(value=0.0, unit="m3")


VARIABLES = {
    f"{temperature_base.__class__.__name__}": temperature_base.Units.serialize(),
    f"{length_base.__class__.__name__}": length_base.Units.serialize(),
    f"{current_base.__class__.__name__}": current_base.Units.serialize(),
    f"{time_base.__class__.__name__}": time_base.Units.serialize(),
    f"{pressure_base.__class__.__name__}": pressure_base.Units.serialize(),
    f"{mass_base.__class__.__name__}": mass_base.Units.serialize(),
    f"{force_base.__class__.__name__}": force_base.Units.serialize(),
    f"{power_base.__class__.__name__}": power_base.Units.serialize(),
    f"{volumetric_flow_base.__class__.__name__}": volumetric_flow_base.Units.serialize(),
    f"{mass_flow_base.__class__.__name__}": mass_flow_base.Units.serialize(),
    f"{density_base.__class__.__name__}": density_base.Units.serialize(),
    f"{percentage_base.__class__.__name__}": percentage_base.Units.serialize(),
    f"{adimentional_base.__class__.__name__}": adimentional_base.Units.serialize(),
    f"{volume_base.__class__.__name__}": volume_base.Units.serialize()
}

DATATYPES = [
    {'label': 'Float', 'value': 'float'},
    {'label': 'Integer', 'value': 'integer'},
    {'label': 'Boolean', 'value': 'boolean'},
    {'label': 'String', 'value': 'string'}
]



from .core import BaseModel, proxy
from .tags import (
    Manufacturer,
    Segment,
    Tags,
    TagValue,
    Variables,
    Units,
    DataTypes
)
from .alarms import (
    AlarmStates,
    AlarmTypes,
    Alarms,
    AlarmSummary
)

from .opcua import OPCUA
from .users import Roles, Users
from .events import Events
from .logs import Logs
from .machines import Machines, TagsMachines
from .opcua_server import AccessType, OPCUAServer


import unittest
from . import assert_dict_contains_subset
from ..modules.users.users import Users, User
from ..modules.users.roles import roles, Role

USERNAME = "user1"
ROLE_NAME = "admin"
EMAIL = "jhon.doe@gmail.com"
PASSWORD = "123456"
NAME = "Jhon"
LASTNAME = "Doe"

USERNAME2 = "user2"
EMAIL2 = "jhon.doe2@gmail.com"

class TestUsers(unittest.TestCase):

    def setUp(self) -> None:
        
        self.roles = roles
        self.roles._delete_all()
        self.users = Users()
        self.users._delete_all()

        return super().setUp()

    def tearDown(self) -> None:
        delattr(self, "roles")
        delattr(self, "users")
        return super().tearDown()
    
    def test_create_role(self):
        
        admin = Role(name="admin", level=0)
        expected = {
            "name": "admin",
            "level": 0
        }
        assert_dict_contains_subset(expected, admin.serialize())
    
    def test_add_role_to_repo(self):
        
        admin = Role(name="admin", level=0)
        self.roles.add(role=admin)
        self.assertIn(admin, self.roles.roles.values())

    def test_get_role(self):
        
        admin = Role(name="admin", level=0)
        admin_id, _ = self.roles.add(role=admin)
        self.assertEqual(admin, self.roles.get(id=admin_id))

    def test_get_role_by_name(self):
        
        admin = Role(name="admin", level=0)
        self.roles.add(role=admin)
        self.assertEqual(admin, self.roles.get_by_name(name=admin.name))

    def test_get_role_names(self):
        
        roles = ["sudo", "admin", "operator"]
        for role in roles:
            _role = Role(name=role, level=0)
            self.roles.add(role=_role)

        self.assertListEqual(roles, self.roles.get_names())

    def test_put_role(self):
        
        role = Role(name="admin", level=0)
        role_id, _ = self.roles.add(role=role)
        self.roles.put(id=role_id, name="sudo")
        self.assertEqual(role.name, "sudo")

    def test_delete_role(self):
        
        roles = ["sudo", "admin", "operator"]
        for role in roles:
            _role = Role(name=role, level=0)
            role_id = self.roles.add(role=_role)
        role = self.roles.get(id=role_id)
        self.roles.delete(id=role_id)
        self.assertNotIn(role, self.roles.roles.values())

    def test_signup(self):

        role = Role(name=ROLE_NAME, level=0)
        self.roles.add(role=role)
        user, _ = self.users.signup(username=USERNAME, role_name=ROLE_NAME, email=EMAIL, password=PASSWORD, name=NAME, lastname=LASTNAME)
        self.assertIsInstance(user, User)

    def test_login_logout(self):

        role = Role(name=ROLE_NAME, level=0)
        self.roles.add(role=role)
        user, _ = self.users.signup(username=USERNAME, role_name=ROLE_NAME, email=EMAIL, password=PASSWORD, name=NAME, lastname=LASTNAME)
        with self.subTest("Test Login"):
            
            self.assertTrue(self.users.login(password="123456", username="user1"))

        with self.subTest("Test Active user"):

            self.assertEqual(user, self.users.get_active_user(token=user.token))

        with self.subTest("Test Logout"):

            self.users.logout(user.token)
            self.assertIsNone(self.users.get_active_user(token=user.token))

    def test_login_fail(self):

        role = Role(name=ROLE_NAME, level=0)
        self.roles.add(role=role)
        user, _ = self.users.signup(username=USERNAME, role_name=ROLE_NAME, email=EMAIL, password=PASSWORD, name=NAME, lastname=LASTNAME)
        self.assertFalse(self.users.login(password=user.password, username="user1"))

    def test_get_user(self):

        role = Role(name=ROLE_NAME, level=0)
        self.roles.add(role=role)
        user, _ = self.users.signup(username=USERNAME, role_name=ROLE_NAME, email=EMAIL, password=PASSWORD, name=NAME, lastname=LASTNAME)
        self.assertEqual(user, self.users.get(identifier=user.identifier))

    def test_get_by_username(self):

        role = Role(name=ROLE_NAME, level=0)
        self.roles.add(role=role)
        user, _ = self.users.signup(username=USERNAME, role_name=ROLE_NAME, email=EMAIL, password=PASSWORD, name=NAME, lastname=LASTNAME)
        self.assertEqual(user, self.users.get_by_username(username=USERNAME))

    def test_get_by_email(self):

        role = Role(name=ROLE_NAME, level=0)
        self.roles.add(role=role)
        user, _ = self.users.signup(username=USERNAME, role_name=ROLE_NAME, email=EMAIL, password=PASSWORD, name=NAME, lastname=LASTNAME)
        self.assertEqual(user, self.users.get_by_email(email=EMAIL))

    def test_get_active_user(self):

        role = Role(name=ROLE_NAME, level=0)
        self.roles.add(role=role)
        user1, _ = self.users.signup(username=USERNAME, role_name=ROLE_NAME, email=EMAIL, password=PASSWORD, name=NAME, lastname=LASTNAME)
        self.users.signup(username=USERNAME2, role_name=ROLE_NAME, email=EMAIL2, password=PASSWORD, name=NAME, lastname=LASTNAME)
        self.users.login(password=PASSWORD, username=USERNAME)
        self.assertEqual(user1, self.users.get_active_user(token=user1.token))

    def test_get_not_active_user(self):

        role = Role(name=ROLE_NAME, level=0)
        self.roles.add(role=role)
        self.users.signup(username=USERNAME, role_name=ROLE_NAME, email=EMAIL, password=PASSWORD, name=NAME, lastname=LASTNAME)
        user2, _ = self.users.signup(username=USERNAME2, role_name=ROLE_NAME, email=EMAIL2, password=PASSWORD, name=NAME, lastname=LASTNAME)
        self.users.login(password=PASSWORD, username=USERNAME)
        self.assertIsNone(self.users.get_active_user(token=user2.token))




# -*- coding: utf-8 -*-
"""automation/logger/datalogger.py

This module implements the Data Logger, responsible for persisting tag values (time-series data)
and managing tag configurations in the database.
"""
import pytz, logging, math
from peewee import fn
from collections import defaultdict
from datetime import datetime
from ..tags.tag import Tag
from ..dbmodels import Tags, TagValue, Units, Segment, Variables
from ..modules.users.users import User
from ..tags.cvt import CVTEngine
from .core import BaseLogger, BaseEngine
from ..variables import *
from ..utils.decorators import db_rollback


DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S.%f"

class DataLogger(BaseLogger):
    """
    Data Logger class.

    This class serves as an API for managing tag settings and accessing logged historical data.
    It interacts directly with the database models.
    """

    def __init__(self):

        super(DataLogger, self).__init__()
        self.tag_engine = CVTEngine()

    @db_rollback
    def set_tag(
        self, 
        id:str,
        name:str, 
        unit:str, 
        data_type:str, 
        description:str="", 
        display_name:str="",
        display_unit:str=None,
        opcua_address:str=None, 
        node_namespace:str=None,
        scan_time:int=None,
        dead_band:float=None,
        manufacturer:str="",
        segment:str=""
        ):
        r"""
        Creates a new tag definition in the database.

        **Parameters:**

        * **id** (str): Unique identifier for the tag.
        * **name** (str): Tag name.
        * **unit** (str): Measurement unit.
        * **data_type** (str): Data type (float, int, bool, etc.).
        * **description** (str, optional): Tag description.
        * **display_name** (str, optional): Friendly name for display.
        * **display_unit** (str, optional): Unit for display purposes.
        * **opcua_address** (str, optional): Source OPC UA server address.
        * **node_namespace** (str, optional): Source OPC UA node ID.
        * **scan_time** (int, optional): Scan interval in ms.
        * **dead_band** (float, optional): Deadband for logging.
        * **manufacturer** (str, optional): Associated manufacturer.
        * **segment** (str, optional): Associated segment.
        """
        if not self.check_connectivity():

            return None
            
        Tags.create(
            id=id,
            name=name, 
            unit=unit,
            data_type=data_type,
            description=description,
            display_name=display_name,
            display_unit=display_unit,
            opcua_address=opcua_address,
            node_namespace=node_namespace,
            scan_time=scan_time,
            dead_band=dead_band,
            manufacturer=manufacturer,
            segment=segment
            )
            
    @db_rollback
    def delete_tag(self, id:str):
        r"""
        Deactivates a tag in the database (logical delete).

        **Parameters:**

        * **id** (str): Tag ID.
        """
        if not self.check_connectivity():

            return None
        
        tag, _ = Tags.get_or_create(identifier=id)
        Tags.put(id=tag.id, active=False)

    @db_rollback
    def get_tag_by_name(self, name:str):
        r"""
        Retrieves a tag configuration by its name.

        **Parameters:**

        * **name** (str): Tag name.

        **Returns:**

        * **Tags**: The tag model instance.
        """
        if not self.check_connectivity():

            return None
        
        return Tags.read_by_name(name=name)

    @db_rollback
    def update_tag(self, id:str, **kwargs):
        r"""
        Updates tag configuration properties.

        **Parameters:**

        * **id** (str): Tag ID.
        * **kwargs**: Fields to update (e.g., gaussian_filter=True).
        """
        if not self.check_connectivity():

            return None
        
        tag = Tags.get(identifier=id)

        if "gaussian_filter" in kwargs:
            
            if kwargs['gaussian_filter'].lower() in ('1', 'true'):

                kwargs['gaussian_filter'] = True

            else:

                kwargs['gaussian_filter'] = False
        
        return Tags.put(id=tag.id, **kwargs)

    @db_rollback
    def set_tags(self, tags):
        r"""
        Batch creates multiple tags.

        **Parameters:**

        * **tags** (list): List of tag dictionaries.
        """
        if not self.check_connectivity():

            return None
        
        for tag in tags:

            self.set_tag(tag)

    @db_rollback
    def get_tags(self):
        r"""
        Retrieves all tags configured in the database.

        **Returns:**

        * **list**: List of all Tags.
        """
        if not self.check_connectivity():

            return list()
            
        return Tags.read_all()
    
    @db_rollback
    def write_tag(self, tag, value, timestamp):
        r"""
        Writes a single value for a tag to the historical log.

        **Parameters:**

        * **tag** (str): Tag name.
        * **value** (float): Value to log.
        * **timestamp** (datetime): Timestamp of the value.
        """
        if not self.is_history_logged:

            return None
        
        if not self.check_connectivity():

            return None
        
        trend = Tags.read_by_name(tag)
        unit = Units.read_by_unit(unit=trend.display_unit.unit)
        TagValue.create(tag=trend, value=value, timestamp=timestamp, unit=unit)

    @db_rollback
    def write_tags(self, tags:list):
        r"""
        Batch writes multiple tag values to the historical log.

        **Parameters:**

        * **tags** (list): List of dictionaries containing {'tag': name, 'value': val, 'timestamp': ts}.
        """
        if not self.is_history_logged:

            return None
        
        if not self.check_connectivity():

            return None
            
        _tags = tags.copy()
        
        for counter, tag in enumerate(tags):
            
            _tag = Tags.read_by_name(tag['tag'])
            
            if _tag:

                unit = Units.get_or_none(id=_tag.display_unit.id)
                _tags[counter].update({
                    'tag': _tag,
                    'unit': unit
                })
        
        TagValue.insert_many(_tags).execute()

    @db_rollback
    def read_trends(self, start:str, stop:str, timezone:str, tags):
        r"""
        Reads historical data for charting/trending.
        
        Supports automatic downsampling based on the requested time span:
        - > 1 week: Daily averages
        - > 2 days: Hourly averages
        - > 2 hours: Minute averages
        - Otherwise: Raw data

        **Parameters:**

        * **start** (str): Start datetime string.
        * **stop** (str): End datetime string.
        * **timezone** (str): Timezone for the query.
        * **tags** (list): List of tag names to query.

        **Returns:**

        * **dict**: Dictionary of time-series data per tag.
        """  
        
        if not self.is_history_logged:

            return None
        
        if not self.check_connectivity():

            return dict()

        _timezone = pytz.timezone(timezone)
        start = _timezone.localize(datetime.strptime(start, DATETIME_FORMAT)).astimezone(pytz.UTC).timestamp()
        stop = _timezone.localize(datetime.strptime(stop, DATETIME_FORMAT)).astimezone(pytz.UTC).timestamp()  
              
        query = (TagValue
                .select(Tags.name, TagValue.value, TagValue.timestamp,
                        Units.unit.alias('tag_value_unit'), Variables.name.alias('variable_name'))
                .join(Tags)
                .join(Units, on=(Tags.unit == Units.id))
                .join(Variables, on=(Units.variable_id == Variables.id))
                .where((TagValue.timestamp.between(start, stop)) & (Tags.name.in_(tags)))
                .order_by(TagValue.timestamp)
                .dicts()) 
         
        # Structure the data
        time_span = (stop - start ) / 60 # span in minutes
        result = defaultdict(lambda: {"values": []})
        if time_span > 60 * 24 * 7:  # 1 week
            # Aggregate data every 1 day
            result = self._agregate_data_every_seconds(query=query, result=result, seconds=3600 * 24, timezone=timezone)

        elif time_span > 60 * 24 * 2:  # 2 days
            # Aggregate data every 1 hora
            result = self._agregate_data_every_seconds(query=query, result=result, seconds=3600, timezone=timezone)

        elif time_span > 60 * 2:  # 2 horas
            # Aggregate data every 1 minute
            result = self._agregate_data_every_seconds(query=query, result=result, seconds=60, timezone=timezone)

        else:
            # Use original data
            
            for entry in query:
                
                from_timezone = pytz.timezone('UTC')
                timestamp = entry['timestamp']
                timestamp = from_timezone.localize(timestamp)
                result[entry['name']]["values"].append({
                    "x": timestamp.astimezone(_timezone).strftime(self.tag_engine.DATETIME_FORMAT),
                    "y": entry['value']
                })
        
        for tag in tags:

            result[tag]['unit'] = self.tag_engine.get_display_unit_by_tag(tag)
        
        return result

    @db_rollback
    def read_table(self, start:str, stop:str, timezone:str, tags:list, page:int=1, limit:int=20):
        r"""
        Retrieves historical data in a paginated table format.

        **Parameters:**

        * **start** (str): Start datetime string.
        * **stop** (str): Stop datetime string.
        * **timezone** (str): Timezone.
        * **tags** (list): List of tag names.
        * **page** (int): Page number.
        * **limit** (int): Records per page.

        **Returns:**

        * **dict**: {data: list, pagination: dict}
        """
        if not self.is_history_logged:
            return None
        
        if not self.check_connectivity():
            return dict()

        _timezone = pytz.timezone(timezone)
        try:
            start_dt = _timezone.localize(datetime.strptime(start, DATETIME_FORMAT)).astimezone(pytz.UTC).timestamp()
            stop_dt = _timezone.localize(datetime.strptime(stop, DATETIME_FORMAT)).astimezone(pytz.UTC).timestamp()
        except ValueError:
            return dict()

        # Base query
        query = (TagValue
                .select(Tags.name, TagValue.value, TagValue.timestamp,
                        Units.unit.alias('tag_value_unit'))
                .join(Tags)
                .join(Units, on=(Tags.unit == Units.id))
                .where((TagValue.timestamp.between(start_dt, stop_dt)) & (Tags.name.in_(tags)))
                .order_by(TagValue.timestamp.desc()))

        total_records = query.count()
        
        # Safe pagination
        if limit <= 0: limit = 20
        if page <= 0: page = 1
        
        total_pages = math.ceil(total_records / limit)
        if total_pages == 0: total_pages = 1
        
        has_next = page < total_pages
        has_prev = page > 1

        paginated_query = query.paginate(page, limit).dicts()
        
        data = []
        utc_timezone = pytz.timezone('UTC')
        
        for entry in paginated_query:
            ts_val = entry['timestamp']
            if isinstance(ts_val, (int, float)):
                dt_object = datetime.fromtimestamp(ts_val, pytz.UTC)
            else:
                # Assuming naive datetime in UTC (based on read_trends using 'UTC' timezone to localize)
                dt_object = utc_timezone.localize(ts_val) if ts_val.tzinfo is None else ts_val
                
            formatted_ts = dt_object.astimezone(_timezone).strftime(DATETIME_FORMAT)
            
            data.append({
                "timestamp": formatted_ts,
                "tag_name": entry['name'],
                "value": f"{entry['value']} {entry['tag_value_unit']}"
            })

        return {
            "data": data,
            "pagination": {
                "page": page,
                "limit": limit,
                "total_records": total_records,
                "total_pages": total_pages,
                "has_next": has_next,
                "has_prev": has_prev
            }
        }
        
    @db_rollback
    def read_tabular_data(self, start:str, stop:str, timezone:str, tags:list, sample_time:int, page:int=1, limit:int=20):
        r"""
        Retrieves historical data in a tabular format with forward-fill resampling.
        
        Useful for exporting data where all tags need to share the same timestamps.

        **Parameters:**

        * **start** (str): Start datetime string.
        * **stop** (str): Stop datetime string.
        * **timezone** (str): Timezone.
        * **tags** (list): List of tag names.
        * **sample_time** (int): Resampling interval in seconds.
        * **page** (int): Page number.
        * **limit** (int): Rows per page.
            
        **Returns:**

        * **dict**: {tag_names, display_names, values, pagination}
        """
        if not self.is_history_logged:
            return None
        
        if not self.check_connectivity():
            return dict()

        _timezone = pytz.timezone(timezone)
        utc_timezone = pytz.UTC
        
        try:
            start_dt = _timezone.localize(datetime.strptime(start, DATETIME_FORMAT)).astimezone(utc_timezone)
            stop_dt = _timezone.localize(datetime.strptime(stop, DATETIME_FORMAT)).astimezone(utc_timezone)
            start_ts = start_dt.timestamp()
            stop_ts = stop_dt.timestamp()
        except ValueError:
            return dict()

        if sample_time <= 0:
            return dict()

        # Get display names mapping early
        try:
            tags_info = Tags.select(Tags.name, Tags.display_name).where(Tags.name.in_(tags)).dicts()
            display_map = {t['name']: t['display_name'] for t in tags_info}
        except Exception:
            display_map = {}

        tag_names = ["timestamp"] + tags
        display_names = ["timestamp"] + [display_map.get(tag, tag) for tag in tags]

        empty_result = {
            "tag_names": tag_names,
            "display_names": display_names,
            "values": [],
            "pagination": {}
        }

        # Check for data presence to adjust start time if necessary
        # 1. Check if there is any data BEFORE or AT start_dt (history)
        has_history = (TagValue
            .select()
            .join(Tags)
            .where(
                (Tags.name.in_(tags)) & 
                (TagValue.timestamp <= start_dt)
            )
            .limit(1)
            .count() > 0)
        
        if not has_history:
            # 2. If no history, find the first actual data point within the requested range
            min_ts = (TagValue
                .select(fn.Min(TagValue.timestamp))
                .join(Tags)
                .where(
                    (Tags.name.in_(tags)) & 
                    (TagValue.timestamp >= start_dt) & 
                    (TagValue.timestamp <= stop_dt) &
                    (TagValue.value.is_null(False))
                )
                .scalar())
            
            if min_ts is None:
                # No data in range and no history
                return empty_result
            
            # Adjust start to the first actual data point
            if isinstance(min_ts, datetime):
                 if min_ts.tzinfo is None:
                     min_ts = utc_timezone.localize(min_ts)
                 start_dt = min_ts
                 start_ts = start_dt.timestamp()
            elif isinstance(min_ts, (int, float)):
                 start_ts = float(min_ts)
                 start_dt = datetime.fromtimestamp(start_ts, pytz.UTC)

        # Calculate total records based on time range and sample time
        total_duration = stop_ts - start_ts
        if total_duration < 0:
            return empty_result
            
        total_records = math.floor(total_duration / sample_time) + 1
        
        # Pagination calculations
        if limit <= 0: limit = 20
        if page <= 0: page = 1
        
        total_pages = math.ceil(total_records / limit)
        if total_pages == 0: total_pages = 1
        
        has_next = page < total_pages
        has_prev = page > 1
        
        # Calculate start and end for current page
        start_index = (page - 1) * limit
        end_index = min(start_index + limit, total_records)
        
        page_start_ts = start_ts + (start_index * sample_time)
        page_end_ts = start_ts + ((end_index - 1) * sample_time)
        
        # Query data needed for this page plus context for forward fill
        
        data_points = []
        current_ts = page_start_ts
        
        # 1. Get initial values (state at page_start_ts)
        current_values = {}
        current_dt = datetime.fromtimestamp(current_ts, pytz.UTC)
        
        for tag_name in tags:
            # Get the latest value <= current_dt
            last_val_query = (TagValue
                .select(TagValue.value)
                .join(Tags)
                .where((Tags.name == tag_name) & (TagValue.timestamp <= current_dt))
                .order_by(TagValue.timestamp.desc())
                .limit(1)
                .dicts())
            
            entry = list(last_val_query)
            if entry:
                current_values[tag_name] = entry[0]['value']
            else:
                current_values[tag_name] = None

        # 2. Get changes within the page window
        page_start_dt = datetime.fromtimestamp(page_start_ts, pytz.UTC)
        page_end_dt = datetime.fromtimestamp(page_end_ts, pytz.UTC)

        changes_query = (TagValue
            .select(Tags.name, TagValue.value, TagValue.timestamp)
            .join(Tags)
            .where(
                (Tags.name.in_(tags)) & 
                (TagValue.timestamp > page_start_dt) & 
                (TagValue.timestamp <= page_end_dt) &
                (TagValue.value.is_null(False))
            )
            .order_by(TagValue.timestamp.asc())
            .dicts())
            
        # Organize changes by timestamp
        changes_by_ts = defaultdict(dict)
        for change in changes_query:
            ts_val = change['timestamp']
            if isinstance(ts_val, datetime):
                if ts_val.tzinfo is None:
                    ts_val = utc_timezone.localize(ts_val)
                ts = ts_val.timestamp()
            else:
                ts = float(ts_val)
            
            changes_by_ts[ts][change['name']] = change['value']
            
        # 3. Generate tabular data
        changes_iter = sorted(changes_by_ts.keys())
        change_idx = 0
        
        for i in range(end_index - start_index):
            step_ts = page_start_ts + (i * sample_time)
            
            # Update current_values with any changes that happened between last step and now (inclusive)
            while change_idx < len(changes_iter) and changes_iter[change_idx] <= step_ts:
                ts = changes_iter[change_idx]
                for tag, val in changes_by_ts[ts].items():
                    current_values[tag] = val
                change_idx += 1
                
            # Build row
            dt_object = datetime.fromtimestamp(step_ts, pytz.UTC)
            formatted_ts = dt_object.astimezone(_timezone).strftime(DATETIME_FORMAT)
            
            row_values = [formatted_ts]
            has_data = False
            for tag in tags:
                val = current_values.get(tag)
                row_values.append(val)
                if val is not None:
                    has_data = True
                
            if has_data:
                data_points.append(row_values)

        return {
            "tag_names": tag_names,
            "display_names": display_names,
            "values": data_points,
            "pagination": {
                "page": page,
                "limit": limit,
                "total_records": total_records,
                "total_pages": total_pages,
                "has_next": has_next,
                "has_prev": has_prev
            }
        }

    def _agregate_data_every_seconds(self, query, result, seconds:int, timezone:str="UTC"):
        r"""
        Downsamples data by averaging values within specific time buckets.
        
        **Parameters:**

        * **query**: The query object containing raw data.
        * **result**: The result dictionary to append to.
        * **seconds**: Bucket size in seconds.
        * **timezone**: Target timezone for result formatting.
        """
        # Aggregate data
        target_timezone = pytz.timezone(timezone)
        buffer = defaultdict(lambda: {"sum": 0, "count": 0, "last_timestamp": None})

        for entry in query:
            bucket = entry['timestamp'].replace(second=(entry['timestamp'].second // seconds) * seconds, microsecond=0)
            buffer_key = (entry['name'], bucket)
            buffer[buffer_key]["sum"] += entry['value']
            buffer[buffer_key]["count"] += 1
            buffer[buffer_key]["last_timestamp"] = entry['timestamp']
            buffer[buffer_key]['unit'] = entry["tag_value_unit"]
            buffer[buffer_key]['variable'] = entry['variable_name']

        for (tag_name, bucket), data in buffer.items():

            avg_value = data["sum"] / data["count"]
            last_timestamp = data["last_timestamp"]
            from_timezone = pytz.timezone('UTC')
            last_timestamp = from_timezone.localize(last_timestamp)
            result[tag_name]["values"].append({
                "x": last_timestamp.astimezone(target_timezone).strftime(self.tag_engine.DATETIME_FORMAT),
                "y": avg_value
            })
        
        return result
        
    @db_rollback
    def read_segments(self):
        r"""
        Retrieves all configured segments.

        **Returns:**

        * **list**: List of Segment objects.
        """
        if not self.check_connectivity():

            return list()

        return Segment.read_all()


class DataLoggerEngine(BaseEngine):
    r"""
    Thread-safe Engine for the DataLogger.

    This class provides a thread-safe wrapper around the `DataLogger` class,
    ensuring that database operations from multiple threads do not conflict.
    """
    def __init__(self):

        super(DataLoggerEngine, self).__init__()
        self.logger = DataLogger()

    def create_tables(self, tables):
        r"""
        Creates default database tables.

        **Parameters:**

        * **tables** (list): List of database models.
        """
        self.logger.create_tables(tables)

    def drop_tables(self, tables:list):
        r"""
        Drops specified tables from the database.

        **Parameters:**

        * **tables** (list): List of database models.
        """
        self.logger.drop_tables(tables)

    def set_tag(
        self,
        tag:Tag
        ):
        r"""
        Registers a tag for logging in the database, using a thread-safe call.

        **Parameters:**

        * **tag** (Tag): The tag object to register.
        """
        _query = dict()
        _query["action"] = "set_tag"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = tag.id
        _query["parameters"]["name"] = tag.name
        _query["parameters"]["unit"] = tag.unit
        _query["parameters"]["data_type"] = tag.data_type
        _query["parameters"]["description"] = tag.description
        _query["parameters"]["display_name"] = tag.display_name
        _query["parameters"]["display_unit"] = tag.display_unit
        _query["parameters"]["opcua_address"] = tag.opcua_address
        _query["parameters"]["node_namespace"] = tag.node_namespace
        _query["parameters"]["scan_time"] = tag.scan_time
        _query["parameters"]["dead_band"] = tag.dead_band
        _query["parameters"]["manufacturer"] = tag.manufacturer
        _query["parameters"]["segment"] = tag.segment
        
        return self.query(_query)

    def get_tags(self):
        r"""
        Retrieves all tags from the database (thread-safe).
        """
        _query = dict()
        _query["action"] = "get_tags"
        _query["parameters"] = dict()
        
        return self.query(_query)
    
    def get_tag_by_name(self, name:str):
        r"""
        Retrieves a tag by name (thread-safe).
        """
        _query = dict()
        _query["action"] = "get_tag_by_name"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        
        return self.query(_query)
    
    def update_tag(
            self, 
            id:str, 
            user:User|None=None,
            **kwargs
            ):
        r"""
        Updates tag configuration (thread-safe).

        **Parameters:**

        * **id** (str): Tag ID.
        * **kwargs**: Properties to update.
        """

        _query = dict()
        _query["action"] = "update_tag"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        for key, value in kwargs.items():

            _query["parameters"][key] = value
    
        return self.query(_query)
    
    def delete_tag(self, id:str):
        r"""
        Deletes a tag (thread-safe).

        **Parameters:**

        * **id** (str): Tag ID.
        """
        _query = dict()
        _query["action"] = "delete_tag"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        
        return self.query(_query)

    def write_tag(self, tag:str, value:float, timestamp:datetime):
        r"""
        Writes a single tag value (thread-safe).

        **Parameters:**

        * **tag** (str): Tag name.
        * **value** (float): Value.
        * **timestamp** (datetime): Timestamp.
        """
        _query = dict()
        _query["action"] = "write_tag"

        _query["parameters"] = dict()
        _query["parameters"]["tag"] = tag
        _query["parameters"]["value"] = value
        _query["parameters"]["timestamp"] = timestamp

        return self.query(_query)

    def write_tags(self, tags:list):
        r"""
        Batch writes tag values (thread-safe).

        **Parameters:**

        * **tags** (list): List of tag value dictionaries.
        """
        _query = dict()
        _query["action"] = "write_tags"

        _query["parameters"] = dict()
        _query["parameters"]["tags"] = tags

        return self.query(_query)
    
    def read_trends(self, start:str, stop:str, timezone:str, *tags):
        r"""
        Reads trend data (thread-safe).

        **Parameters:**

        * **start** (str): Start time.
        * **stop** (str): End time.
        * **timezone** (str): Timezone.
        * **tags**: Variable length argument of tag names.
        """
        _query = dict()
        _query["action"] = "read_trends"
        _query["parameters"] = dict()
        _query["parameters"]["start"] = start
        _query["parameters"]["stop"] = stop
        _query["parameters"]["timezone"] = timezone
        _query["parameters"]["tags"] = tags
        return self.query(_query)

    def read_tabular_data(self, start:str, stop:str, timezone:str, tags:list, sample_time:int, page:int=1, limit:int=20):
        r"""
        Reads tabular data (thread-safe).
        """
        _query = dict()
        _query["action"] = "read_tabular_data"
        _query["parameters"] = dict()
        _query["parameters"]["start"] = start
        _query["parameters"]["stop"] = stop
        _query["parameters"]["timezone"] = timezone
        _query["parameters"]["tags"] = tags
        _query["parameters"]["sample_time"] = sample_time
        _query["parameters"]["page"] = page
        _query["parameters"]["limit"] = limit
        return self.query(_query)

    def read_segments(self):
        r"""
        Reads segments (thread-safe).
        """
        _query = dict()
        _query["action"] = "read_segments"
        _query["parameters"] = dict()
        return self.query(_query)



from peewee import CharField, IntegerField, ForeignKeyField, FloatField
from .core import BaseModel
from .tags import Tags


class Machines(BaseModel):
    r"""
    Database model for State Machines configuration.
    """
    
    identifier = CharField(unique=True)
    name = CharField(unique=True)
    interval = FloatField()
    threshold = FloatField(null=True)
    on_delay = IntegerField(null=True)
    description = CharField(max_length=128)
    classification = CharField(max_length=128)
    buffer_size = IntegerField()
    buffer_roll_type = CharField(max_length=16)
    criticity = IntegerField()
    priority = IntegerField()

    @classmethod
    def create(
        cls, 
        identifier:str,
        name:str,
        interval:int,
        description:str,
        classification:str,
        buffer_size:int,
        buffer_roll_type:str,
        criticity:int,
        priority:int,
        threshold:float=None,
        on_delay:int=None
        )-> dict:
        r"""
        Creates a new Machine record.

        **Parameters:**

        * **identifier** (str): Unique identifier.
        * **name** (str): Machine name.
        * **interval** (int): Execution interval.
        * **description** (str): Description.
        * **classification** (str): Machine type.
        * **buffer_size** (int): Data buffer size.
        * **buffer_roll_type** (str): Buffer roll strategy.
        * **criticity** (int): Criticity level.
        * **priority** (int): Priority level.

        **Returns:**

        * **dict**: Result status and data.
        """

        result = dict()
        data = dict()

        if not cls.name_exist(name):

            query = cls(
                identifier=identifier,
                name=name,
                interval=interval,
                description=description,
                classification=classification,
                buffer_size=buffer_size,
                buffer_roll_type=buffer_roll_type,
                criticity=criticity,
                priority=priority,
                threshold=threshold,
                on_delay=on_delay
                )
            query.save()
            
            message = f"Machine {name} created successfully"
            data.update(query.serialize())

            result.update(
                {
                    'message': message, 
                    'data': data
                }
            )
            return result

        message = f"Machine {name} is already into database"
        result.update(
            {
                'message': message, 
                'data': data
            }
        )
        return result

    @classmethod
    def read_by_name(cls, name:str)->bool:
        r"""
        Retrieves a machine by name.
        """
        query = cls.get_or_none(name=name)
        
        if query is not None:

            return query
        
        return None

    @classmethod
    def read_config(cls):
        r"""
        Retrieves configuration for all machines.

        **Returns:**

        * **dict**: Map of Machine Name -> Configuration.
        """
        return {f"{query.name}": query.serialize() for query in cls.select()}

    @classmethod
    def name_exist(cls, name:str)->bool:
        r"""
        Checks if a machine name exists.
        """
        query = cls.get_or_none(name=name)
        
        if query is not None:

            return True
        
        return False
    
    def get_tags(self):
        r"""
        Returns related tags.
        """
        return self.tags

    def serialize(self)-> dict:
        r"""
        Serializes the machine record.
        """

        return {
            "id": self.id,
            "identifier": self.identifier,
            "name": self.name,
            "interval": self.interval,
            "description": self.description,
            "classification": self.classification,
            "buffer_size": self.buffer_size,
            "buffer_roll_type": self.buffer_roll_type,
            "criticity": self.criticity,
            "priority": self.priority,
            "threshold": self.threshold,
            "on_delay": self.on_delay
        }


class TagsMachines(BaseModel):
    r"""
    Many-to-Many relationship between Tags and Machines.
    """
    
    tag = ForeignKeyField(Tags, backref="machines")
    machine = ForeignKeyField(Machines, backref="tags")
    default_tag_name = CharField(max_length=64, null=True)

    @classmethod
    def create(
        cls, 
        tag_name:str,
        machine_name:str,
        default_tag_name:str=None
        )-> dict:
        r"""
        Links a Tag to a Machine.

        **Parameters:**

        * **tag_name** (str): Name of the tag.
        * **machine_name** (str): Name of the machine.
        * **default_tag_name** (str, optional): Alias for the tag within the machine context.
        """

        tag = Tags.get_or_none(name=tag_name)
        machine = Machines.get_or_none(name=machine_name)

        if not cls.get_or_none(tag=tag, machine=machine):

            query = cls(
                tag=tag,
                machine=machine,
                default_tag_name=default_tag_name
                )
            query.save()

    def serialize(self):
        r"""
        Serializes the relationship.
        """
        return {
            "machine": self.machine.serialize(),
            "tag": self.tag.serialize(),
            "default_tag_name": self.default_tag_name
        }


from flask import request
from datetime import datetime, timezone
import jwt
import logging
from flask_restx import Namespace, Resource
from .models.users import signup_parser, login_parser, change_password_parser, reset_password_parser, create_tpt_parser
from .... import PyAutomation, TIMEZONE, _TIMEZONE
from ....extensions.api import api
from ....extensions import _api as Api
from ....modules.users.users import Users as CVTUsers
from ....dbmodels.users import Users

DATETIME_FORMAT = "%m/%d/%Y, %H:%M:%S"
ns = Namespace('Users', description='User Management and Authentication')
app = PyAutomation()
users = CVTUsers()

@ns.route('/')
class UsersCollection(Resource):

    @api.doc(security='apikey', description="Retrieves a list of all registered users.")
    @api.response(200, "Success")
    @Api.token_required(auth=True)
    def get(self):
        """
        Get all users.

        Retrieves a list of all users currently registered in the system.
        """

        return users.serialize(), 200

@ns.route('/signup')
class SignUpResource(Resource):
    
    @Api.validate_reqparser(reqparser=signup_parser)
    @api.doc(security=None, description="Registers a new user.")
    @api.response(200, "User created successfully")
    @api.response(400, "User creation failed")
    @ns.expect(signup_parser)
    def post(self):
        """
        User signup.

        Registers a new user with the provided credentials and role.
        """
        args = signup_parser.parse_args()
        user, message = app.signup(**args)
        
        if user:

            return user.serialize(), 200
        
        return message, 400


@ns.route('/login')
class LoginResource(Resource):

    @Api.validate_reqparser(reqparser=login_parser)
    @api.doc(security=None, description="Authenticates a user and returns an API token.")
    @api.response(200, "Login successful")
    @api.response(403, "Invalid credentials")
    @ns.expect(login_parser)
    def post(self):
        """
        User login.

        Authenticates a user using username/email and password. Returns an API key/token.
        """
        args = login_parser.parse_args()
        user, message = app.login(**args)

        if user:

            return {
                "apiKey": user.token,
                "role": user.role.name,
                "role_level": user.role.level,
                "datetime": datetime.now(TIMEZONE).strftime(DATETIME_FORMAT),
                "timezone": _TIMEZONE
                }, 200

        return message, 403
    

@ns.route('/credentials_are_valid')
class VerifyCredentialsResource(Resource):
    
    @api.doc(security='apikey', description="Verifies if the provided credentials are valid without logging in.")
    @api.response(200, "Success (True/False)")
    @Api.token_required(auth=True)
    @Api.validate_reqparser(reqparser=login_parser)
    @ns.expect(login_parser)
    def post(self):
        """
        Verify credentials.

        Checks if the provided username/password combination is valid.
        """
        args = login_parser.parse_args()
        credentials_valid, _ = users.verify_credentials(**args)
        return credentials_valid, 200
    
@ns.route('/<username>')
@api.param('username', 'The username')
class UserResource(Resource):
    
    @api.doc(security='apikey', description="Retrieves information about a specific user.")
    @api.response(200, "Success")
    @api.response(400, "User not found")
    @Api.token_required(auth=True)
    def get(self, username):
        """
        Get user info.

        Retrieves detailed information about a specific user by username.
        """
        
        user = users.get_by_username(username=username)

        if user:

            return user.serialize(), 200

        return f"{username} is not a valid username", 400


@ns.route('/logout')
class LogoutResource(Resource):

    @api.doc(security='apikey', description="Logs out the current user and invalidates the token.")
    @api.response(200, "Logout successful")
    @Api.token_required(auth=True)
    def post(self):
        """
        User logout.

        Invalidates the current session token.
        """
        if 'X-API-KEY' in request.headers:
                            
            token = request.headers['X-API-KEY']

        elif 'Authorization' in request.headers:
            
            token = request.headers['Authorization'].split('Token ')[-1]
        
        _, message = Users.logout(token=token)

        return message, 200

@ns.route('/change_password')
class ChangePasswordResource(Resource):

    @Api.validate_reqparser(reqparser=change_password_parser)
    @api.doc(security='apikey', description="Changes a user's password with role-based authorization.")
    @api.response(200, "Password changed successfully")
    @api.response(400, "Invalid request or authorization denied")
    @api.response(401, "Unauthorized")
    @ns.expect(change_password_parser)
    @Api.token_required(auth=True)
    def post(self):
        """
        Change password.

        Changes a user's password following role-based authorization rules:
        - Sudo users can only change passwords of admin users (not their own)
        - Admin users can change passwords of users with role_level >= admin
        - Other roles can only change their own password
        - When changing own password, current password must be provided and validated
        """
        # Get token from headers
        token = None
        if 'X-API-KEY' in request.headers:
            token = request.headers['X-API-KEY']
        elif 'Authorization' in request.headers:
            token = request.headers['Authorization'].split('Token ')[-1]
        
        if not token:
            return {'message': 'Token is required'}, 401

        # Get current user from token
        current_user = users.get_active_user(token=token)
        if not current_user:
            return {'message': 'Invalid token or user not found'}, 401

        # Parse arguments
        args = change_password_parser.parse_args()
        target_username = args['target_username']
        new_password = args['new_password']
        current_password = args.get('current_password')

        # Get target user
        target_user = users.get_by_username(username=target_username)
        if not target_user:
            return {'message': f'User {target_username} not found'}, 400

        # Extract role information
        current_role_name = current_user.role.name.upper()
        current_role_level = current_user.role.level
        target_role_name = target_user.role.name.upper()
        target_role_level = target_user.role.level

        # Check if changing own password
        is_own_password = current_user.username == target_username

        # Business logic validation - API level restrictions
        if current_role_name == "SUDO":
            # Sudo can only change admin passwords, not own password
            if is_own_password:
                return {'message': 'Sudo users cannot change their own password'}, 400
            if target_role_name != "ADMIN":
                return {'message': 'Sudo users can only change passwords of admin users'}, 400
        elif current_role_name == "ADMIN":
            # Admin can change passwords of users with role_level >= admin
            if target_role_level < current_role_level:
                return {'message': f'Admin users cannot change passwords of users with role level lower than {current_role_level}'}, 400
        else:
            # Other roles can only change their own password
            if not is_own_password:
                return {'message': 'You can only change your own password'}, 400

        # If changing own password, validate current password is provided
        if is_own_password:
            if not current_password:
                return {'message': 'Current password is required when changing your own password'}, 400

        # Call core method (internal, no restrictions)
        result, status_msg = app.change_password(
            target_username=target_username,
            new_password=new_password,
            current_password=current_password
        )

        if result:
            return {'message': status_msg}, 200
        else:
            return {'message': status_msg}, 400

@ns.route('/reset_password')
class ResetPasswordResource(Resource):

    @Api.validate_reqparser(reqparser=reset_password_parser)
    @api.doc(security='apikey', description="Resets a user's password (for forgotten password) with role-based authorization.")
    @api.response(200, "Password reset successfully")
    @api.response(400, "Invalid request or authorization denied")
    @api.response(401, "Unauthorized")
    @ns.expect(reset_password_parser)
    @Api.token_required(auth=True)
    def post(self):
        """
        Reset password.

        Resets a user's password when forgotten, following role-based authorization rules:
        - Sudo users can only reset passwords of admin users (not their own)
        - Admin users can reset passwords of users with role_level >= admin
        - Other roles can only reset their own password
        - No current password validation is required (forgotten password scenario)
        """
        # Get token from headers
        token = None
        if 'X-API-KEY' in request.headers:
            token = request.headers['X-API-KEY']
        elif 'Authorization' in request.headers:
            token = request.headers['Authorization'].split('Token ')[-1]
        
        if not token:
            return {'message': 'Token is required'}, 401

        # Get current user from token
        current_user = users.get_active_user(token=token)
        if not current_user:
            return {'message': 'Invalid token or user not found'}, 401

        # Parse arguments
        args = reset_password_parser.parse_args()
        target_username = args['target_username']
        new_password = args['new_password']

        # Get target user
        target_user = users.get_by_username(username=target_username)
        if not target_user:
            return {'message': f'User {target_username} not found'}, 400

        # Extract role information
        current_role_name = current_user.role.name.upper()
        current_role_level = current_user.role.level
        target_role_name = target_user.role.name.upper()
        target_role_level = target_user.role.level

        # Check if resetting own password
        is_own_password = current_user.username == target_username

        # Business logic validation - API level restrictions
        if current_role_name == "SUDO":
            # Sudo can only reset admin passwords, not own password
            if is_own_password:
                return {'message': 'Sudo users cannot reset their own password'}, 400
            if target_role_name != "ADMIN":
                return {'message': 'Sudo users can only reset passwords of admin users'}, 400
        elif current_role_name == "ADMIN":
            # Admin can reset passwords of users with role_level >= admin
            if target_role_level < current_role_level:
                return {'message': f'Admin users cannot reset passwords of users with role level lower than {current_role_level}'}, 400
        else:
            # Other roles can only reset their own password
            if not is_own_password:
                return {'message': 'You can only reset your own password'}, 400

        # Call core method (internal, no restrictions, no current password validation)
        result, status_msg = app.reset_password(
            target_username=target_username,
            new_password=new_password
        )

        if result:
            return {'message': status_msg}, 200
        else:
            return {'message': status_msg}, 400

@ns.route('/create_tpt')
class CreateTPTResource(Resource):

    @Api.validate_reqparser(reqparser=create_tpt_parser)
    @api.doc(security='apikey', description="Creates a Third Party Token (TPT) JWT for a specific role. Only sudo users can access this endpoint.")
    @api.response(200, "Token created successfully")
    @api.response(400, "Invalid request")
    @api.response(401, "Unauthorized")
    @api.response(403, "Forbidden - Sudo access required")
    @ns.expect(create_tpt_parser)
    @Api.token_required(auth=True)
    @Api.auth_roles(['sudo'])
    def post(self):
        """
        Create Third Party Token (TPT).

        Creates a JWT token for third-party integration with a specific role embedded.
        This endpoint is restricted to sudo users only.
        
        The token is signed using AUTOMATION_APP_SECRET_KEY and can be used
        for third-party API authentication.
        """
        # Parse arguments
        args = create_tpt_parser.parse_args()
        role_name = args['role_name']

        try:
            # Create JWT payload
            payload = {
                "created_on": datetime.now(timezone.utc).strftime(app.cvt.DATETIME_FORMAT),
                "role": role_name
            }
            
            # Encode token using AUTOMATION_APP_SECRET_KEY
            token = jwt.encode(
                payload, 
                app.server.config['AUTOMATION_APP_SECRET_KEY'], 
                algorithm="HS256"
            )
            
            return {
                'token': token,
                'role': role_name,
                'created_on': payload['created_on'],
                'message': 'Third Party Token created successfully'
            }, 200
            
        except Exception as err:
            logger = logging.getLogger("pyautomation")
            logger.error(f"Error creating TPT: {str(err)}")
            return {'message': f'Error creating token: {str(err)}'}, 500


import pytz
from peewee import CharField, TimestampField, ForeignKeyField, fn
from ..dbmodels.core import BaseModel
from datetime import datetime
from .users import Users
from .events import Events
from .alarms import AlarmSummary, Alarms
from ..modules.users.users import User

DATETIME_FORMAT = "%m/%d/%Y, %H:%M:%S.%f"


class Logs(BaseModel):
    r"""
    Database model for Application Logs.

    Logs store system messages, errors, and informational records, optionally linked to
    events or alarms.
    """
    
    timestamp = TimestampField(utc=True)
    message = CharField(max_length=256)
    description = CharField(max_length=256, null=True)
    classification = CharField(max_length=128, null=True)
    user = ForeignKeyField(Users, backref='logs', on_delete='CASCADE')
    alarm = ForeignKeyField(AlarmSummary, null=True, backref='logs', on_delete='CASCADE')
    event = ForeignKeyField(Events, null=True, backref='logs', on_delete='CASCADE')

    @classmethod
    def create(
        cls, 
        message:str, 
        user:User, 
        description:str=None, 
        classification:str=None,
        alarm_summary_id:int=None,
        event_id:int=None,
        timestamp:datetime=None
        )->tuple:
        r"""
        Creates a new log entry.

        **Parameters:**

        * **message** (str): Log content.
        * **user** (User): User creating the log.
        * **description** (str, optional): Additional details.
        * **classification** (str, optional): Log type/category.
        * **alarm_summary_id** (int, optional): Link to an alarm summary entry.
        * **event_id** (int, optional): Link to an event.
        * **timestamp** (datetime, optional): Log time.

        **Returns:**

        * **tuple**: (Query object, status message)
        """
        if not isinstance(user, User):

            return None, f"User {user} - {type(user)} must be an User Object"
        
        _user = Users.read_by_username(username=user.username) 

        if not timestamp:

            timestamp = datetime.now()
        
        if not isinstance(timestamp, datetime):

            return None, f"Timestamp must be a datetime Object"
        
        query = cls(
            message=message,
            user=_user,
            description=description,
            classification=classification,
            timestamp=timestamp,
            event=Events.get_or_none(id=event_id),
            alarm=AlarmSummary.get_or_none(id=alarm_summary_id)
        )
        query.save()

        return query, f"Event creation successful"
    
    @classmethod
    def read_lasts(cls, lasts:int=1):
        r"""
        Retrieves the last N logs.

        **Parameters:**

        * **lasts** (int): Number of logs to retrieve.

        **Returns:**

        * **list**: List of serialized log dictionaries.
        """
        logs = cls.select().order_by(cls.id.desc()).limit(lasts)

        return [log.serialize() for log in logs]
    
    @classmethod
    def filter_by(
        cls, 
        usernames:list[str]=None,
        alarm_names:list[str]=None,
        event_ids:list[int]=None,
        description:str="",
        message:str="",
        classification:str="",
        greater_than_timestamp:datetime=None,
        less_than_timestamp:datetime=None,
        timezone:str='UTC'
        ):
        r"""
        Filters logs based on various criteria.

        **Parameters:**

        * **usernames** (list[str]): Filter by user.
        * **alarm_names** (list[str]): Filter by linked alarm name.
        * **event_ids** (list[int]): Filter by linked event ID.
        * **message**, **description**, **classification**: Text search.
        * **greater_than_timestamp**, **less_than_timestamp**: Time range.

        **Returns:**

        * **list**: List of matching logs.
        """
        _timezone = pytz.timezone(timezone)
        query = cls.select()
        
        if usernames:
            subquery = Users.select(Users.id).where(Users.username.in_(usernames))
            query = query.join(Users).where(Users.id.in_(subquery))
        
        if event_ids:
            subquery = Events.select(Events.id).where(Events.id.in_(event_ids))
            query = query.join(Events).where(Events.id.in_(subquery))
        
        if alarm_names:
            subquery = Alarms.select(Alarms.id).where(Alarms.name.in_(alarm_names))
            alarm_subquery = AlarmSummary.select(AlarmSummary.id).join(Alarms).where(Alarms.id.in_(subquery))
            query = query.join(AlarmSummary).where(AlarmSummary.id.in_(alarm_subquery))
        
        if description:
            query = query.where(fn.LOWER(cls.description).contains(description.lower()))
        
        if message:
            query = query.where(fn.LOWER(cls.message).contains(message.lower()))
        
        if classification:
            query = query.where(fn.LOWER(cls.classification).contains(classification.lower()))
        
        if greater_than_timestamp:
            greater_than_timestamp = _timezone.localize(datetime.strptime(greater_than_timestamp, '%Y-%m-%d %H:%M:%S.%f')).astimezone(pytz.UTC)
            query = query.where(cls.timestamp > greater_than_timestamp)
        
        if less_than_timestamp:
            less_than_timestamp = _timezone.localize(datetime.strptime(less_than_timestamp, '%Y-%m-%d %H:%M:%S.%f')).astimezone(pytz.UTC)
            query = query.where(cls.timestamp < less_than_timestamp)
        
        query = query.order_by(cls.id.desc())

        if not query.exists():
            
            return []
        
        return [log.serialize() for log in query]

    def serialize(self)-> dict:
        r"""
        Serializes the log record.
        """
        from .. import MANUFACTURER, SEGMENT, TIMEZONE
        timestamp = self.timestamp
        if timestamp:
            timestamp = pytz.UTC.localize(timestamp).astimezone(TIMEZONE)
            timestamp = timestamp.strftime(DATETIME_FORMAT)

        _event = None
        if self.event:

            _event = self.event.serialize()

        _alarm = None
        if self.alarm:

            _alarm = self.alarm.serialize()

        return {
            "id": self.id,
            "timestamp": timestamp,
            "user": self.user.serialize(),
            "message": self.message,
            "description": self.description,
            "classification": self.classification,
            "event": _event,
            "alarm": _alarm,
            "segment": SEGMENT,
            "manufacturer": MANUFACTURER
        }


from ..utils.decorators import decorator
from ..buffer import Buffer
import math

data = dict()

def __iad(data:Buffer, tag_name:str):
    r"""
    Frozen Data Algorithm
    """
    from ..managers.alarms import AlarmManager
    alarm_manager = AlarmManager()
    alarm = alarm_manager.get_alarm_by_name(name=f"alarm.iad.{tag_name}")

    if alarm:
        if alarm.state.alarm_status.lower() == "not active":

            mean = sum(data) / len(data)
            variance = sum((x - mean) ** 2 for x in data) / len(data)
            std_dev = math.sqrt(variance)
            
            if abs(std_dev) < 0.001:
                
                alarm.description = f"Frozen data anomaly"
                alarm.abnormal_condition()
            
            else:
                
                alarm.description = ""
                alarm.normal_condition()

@decorator
def iad_frozen_data(func, args, kwargs):
    r"""
    Documentation here
    """
    
    cvt = args[0]
    tag_id = kwargs["id"]
    value = kwargs["value"]
    tag = cvt.get_tag(id=tag_id)
    if tag.frozen_data_detection:
        
        if tag.name not in data:
            
            data[tag.name] = Buffer()

        data[tag.name](value)
        # Apply IAD logic
        if len(data[tag.name]) >= data[tag.name].size:
            
            __iad(data[tag.name], tag_name=tag.name)

    return func(*args, **kwargs)


import dash
import dash_bootstrap_components as dbc
from automation.models import StringType, IntegerType, FloatType
from ..components.machines import MachinesComponents
from ...variables import *

def init_callback(app:dash.Dash):

    @app.callback(
        dash.Output('machines_detailed_tabs', 'children'),
        dash.Output('machines_detailed_tabs', 'active_tab'),
        dash.Output("machine_detailed_content", "children"),
        dash.Input('machines_detailed_page', 'pathname'),
        prevent_initial_call=True
        )
    def display_page(pathname):
        r"""
        Documentation here
        """
        tabs = list()
        active_tab = None
        machine_tab_content = dash.html.P("This shouldn't ever be displayed...")
        if pathname=="/machines-detailed":
            for machine, _, _ in app.automation.get_machines():

                if hasattr(machine, "classification"):

                    if machine.classification.value.lower()=="opc ua server":

                        continue

                    elif machine.classification.value.lower()=="data acquisition system":

                        continue

                    internal_variables = machine.get_internal_process_type_variables()
                    if internal_variables:

                        tabs.append(dbc.Tab(label=f"{machine.name.value}", tab_id=f"tab-{machine.name.value}"))
                        if not active_tab:

                            active_tab = f"tab-{machine.name.value}"
                            machine_tab_content = MachinesComponents.machine_tab_content(app=app.automation, machine_name=machine.name.value)
        
        return tabs, active_tab, machine_tab_content
    
    @app.callback(
        dash.Output('machines_detailed_datatable', 'data', allow_duplicate=True),
        dash.Input('timestamp-interval', 'n_intervals'),
        dash.State('machines_detailed_tabs', 'active_tab'),
        prevent_initial_call=False
    )
    def update_table( n_intervals, active_tab:str):
        machine_name = active_tab.split("-")[-1]
        def flatten_dict(d, parent_key='', sep='.'): 
            items = [] 
            for k, v in d.items(): 
                
                new_key = f"{parent_key}{sep}{k}" if parent_key else k 
                if isinstance(v, dict): 
                    items.extend(flatten_dict(v, new_key, sep=sep).items()) 
                else: 
                    items.append((new_key, v)) 
                    
            return dict(items)

        machine = app.automation.get_machine(name=StringType(machine_name))
        data = list()
        serialized_data = machine.serialize()
        flattened_data = flatten_dict(serialized_data)
        for key, value in flattened_data.items():

            if isinstance(value, (bool, int, str, float)):

                data.append({"name": key, "value": value})

        return data
    
    @app.callback(
        dash.Output('subscribed_tag_machine_input', 'options', allow_duplicate=True),
        dash.Output('field_tag_input', 'options', allow_duplicate=True),
        dash.Output('internal_tag_input', 'options', allow_duplicate=True),
        dash.Input('machines_detailed_tabs', 'active_tab'),
        prevent_initial_call=False
    )
    def update_subscription_form(active_tab:str):
        machine_name = active_tab.split("-")[-1]
        machine = app.automation.get_machine(name=StringType(machine_name))
        internal_variables = machine.get_read_only_process_type_variables()
        subscribed_tags_machine = [{"label": "", "value": ""}]
        not_subscribed = [{"label": "", "value": ""}]
        available_tags = [{"label": "", "value": ""}]
        tags = app.automation.cvt._cvt.get_cuasi_field_tags_names()
        for _tag_name, value in internal_variables.items():
            
            if value.tag:
                subscribed_tags_machine.append({
                    "label": f"{value.tag.name}->{_tag_name}", "value": value.tag.name
                    })
                if value.tag.name in tags:
                    tags.remove(value.tag.name)
                
            else:

                not_subscribed.append({
                    "label": _tag_name, "value": _tag_name
                })

        for tag in tags:

            available_tags.append({
                "label": tag, "value": tag
            })
        return subscribed_tags_machine, available_tags, not_subscribed
    
    @app.callback(
        dash.Output("machine_threshold_input", "placeholder", allow_duplicate=True),
        dash.Output("machine_threshold_input", "disabled", allow_duplicate=True),
        dash.Output("threshold_input_text", 'children', allow_duplicate=True),
        dash.Output("machine_interval_input", 'placeholder', allow_duplicate=True),
        dash.Output("buffer_size_input", 'placeholder', allow_duplicate=True),
        dash.Output("buffer_size_input", 'disabled', allow_duplicate=True),
        dash.Output("on_delay_input", 'placeholder', allow_duplicate=True),
        dash.Output("machine_threshold_input", "value", allow_duplicate=True),
        dash.Output("machine_interval_input", 'value', allow_duplicate=True),
        dash.Output("buffer_size_input", 'value', allow_duplicate=True),
        dash.Output("on_delay_input", 'value', allow_duplicate=True),
        dash.Input('machines_detailed_tabs', 'active_tab'),
        prevent_initial_call=False
    )
    def update_attributes_form(active_tab:str):
        machine_name = active_tab.split("-")[-1]
        machine = app.automation.get_machine(name=StringType(machine_name))
        machine_interval = machine.get_interval()
        threshold = machine.threshold
        on_delay = machine.on_delay
        machine_threshold_place = f"Current threshold {threshold.value}"
        machine_interval = f"Current machine interval {machine_interval}"
        threshold_input_text = threshold.unit
        buffer_size = f"Buffer Size"
        disable = True
        if hasattr(machine, "buffer_size"):
            disable = False
            buffer_size = machine.buffer_size
            buffer_size = f"Current buffer size: {buffer_size.value}"
        if "pfm" in machine_name.lower():
            disable = True
        elif "observer" in  machine_name.lower():
            disable = True
        on_delay = f"Current on delay: {on_delay.value}"
        
        return machine_threshold_place, disable, threshold_input_text, machine_interval, buffer_size, disable, on_delay, "", "", "", ""
    
    @app.callback(
        dash.Output('machine_state_input', 'children'),
        dash.Input('timestamp-interval', 'n_intervals'),
        dash.State('machines_detailed_tabs', 'active_tab'),
        dash.State('machine_state_input', 'children'),
        prevent_initial_call=False
    )
    def update_state( n_intervals, active_tab:str, previous_machine_state):
        r"""
        Documentation here
        """
        machine_name = active_tab.split("-")[-1]
        machine = app.automation.get_machine(name=StringType(machine_name))
        state = machine.current_state.value

        if state!=previous_machine_state:
            actions = [{"label": "", "value": ""}]
            for action in machine.get_allowed_actions():
                actions.append({"label": action, "value": action})
            dash.set_props("machine_actions_input", {
                'options': actions
                })

            return state

        return previous_machine_state
    
    @app.callback(
        dash.Input("field_tag_input", "value"),
        dash.Input("internal_tag_input", "value"),
        )
    def enable_subscribe_button(
        field_tag:str, 
        internal_tag:str
        )->str:
        r"""
        Documentation here
        """
        if field_tag and internal_tag:

            dash.set_props("subscribe_tag_machine_button", {'disabled': False})

        else:
            
            dash.set_props("subscribe_tag_machine_button", {'disabled': True})

    @app.callback(
        dash.Input("subscribed_tag_machine_input", "value")
        )
    def enable_unsubscribe_button(
        tag_machine:str
        )->str:
        r"""
        Documentation here
        """
        if tag_machine:

            dash.set_props("unsubscribe_tag_machine_button", {'disabled': False})

        else:
            
            dash.set_props("unsubscribe_tag_machine_button", {'disabled': True})

    @app.callback(
        dash.Output("field_tag_input", "value", allow_duplicate=True),
        dash.Output("internal_tag_input", "value", allow_duplicate=True),
        dash.Output('subscribed_tag_machine_input', 'value', allow_duplicate=True),
        dash.Output("field_tag_input", "options", allow_duplicate=True),
        dash.Output("internal_tag_input", "options", allow_duplicate=True),
        dash.Output('subscribed_tag_machine_input', 'options'),
        dash.Input('subscribe_tag_machine_button', 'n_clicks'),
        dash.State("field_tag_input", "value"),
        dash.State("internal_tag_input", "value"),
        dash.State('machines_detailed_tabs', 'active_tab'),
        dash.State('subscribed_tag_machine_input', 'options'),
        dash.State("field_tag_input", "options"),
        dash.State("internal_tag_input", "options"),
        prevent_initial_call=True
    )
    def subscribed_on_click(
        btn1, 
        field_tag_input:str,
        internal_tag_input:str,
        active_tab:str,
        tags_machine_options:list,
        field_tag_options:list,
        internal_tag_options:list,
        ):
        r"""
        Documentation here
        """
        if "subscribe_tag_machine_button" == dash.ctx.triggered_id:

            machine_name = active_tab.split("-")[-1]
            machine = app.automation.get_machine(name=StringType(machine_name))
            field_tag = app.automation.cvt._cvt.get_tag_by_name(name=field_tag_input)

            subscribed, message = machine.subscribe_to(tag=field_tag, default_tag_name=internal_tag_input)

            if subscribed:

                tags_machine_options.append({
                    "label": f"{field_tag_input}->{internal_tag_input}", "value": internal_tag_input
                })
                field_tag_options.remove({
                    "label": field_tag_input, "value": field_tag_input
                })
                internal_tag_options.remove({
                    "label": internal_tag_input, "value": internal_tag_input
                })

            return "", "", "", field_tag_options, internal_tag_options, tags_machine_options

    @app.callback(
        dash.Output('subscribed_tag_machine_input', 'value', allow_duplicate=True),
        dash.Output('subscribed_tag_machine_input', 'options', allow_duplicate=True),
        dash.Output("field_tag_input", "options", allow_duplicate=True),
        dash.Output("internal_tag_input", "options", allow_duplicate=True),
        dash.Input('unsubscribe_tag_machine_button', 'n_clicks'),
        dash.State('subscribed_tag_machine_input', 'value'),
        dash.State('machines_detailed_tabs', 'active_tab'),
        dash.State('subscribed_tag_machine_input', 'options'),
        dash.State("field_tag_input", "options"),
        dash.State("internal_tag_input", "options"),
        prevent_initial_call=True
    )
    def unsubscribed_on_click(
        btn1,
        tag_name:str,
        active_tab:str,
        tags_machine_options:list,
        field_tags_options:list,
        internal_tags_options:list
        ):
        r"""
        Documentation here
        """
        if "unsubscribe_tag_machine_button" == dash.ctx.triggered_id:

            machine_name = active_tab.split("-")[-1]
            machine = app.automation.get_machine(name=StringType(machine_name))
            tag = app.automation.get_tag_by_name(name=tag_name)
            if machine.unsubscribe_to(tag=tag):

                for tag_machine in tags_machine_options:
                    _tag_machine = tag_machine["label"].split("->")
                    _tag_name = _tag_machine[0]
                    _internal_tag = _tag_machine[-1]
                    if _tag_name==tag_name:
                        
                        break
                
                tags_machine_options.remove(tag_machine)
                field_tags_options.append({"label": _tag_name, "value": _tag_name})
                internal_tags_options.append({ "label": _internal_tag, "value": _internal_tag})

            return "", tags_machine_options, field_tags_options, internal_tags_options

    @app.callback(
        dash.Input('machine_actions_input', 'value'),
        dash.State('machines_detailed_tabs', 'active_tab'),
        prevent_initial_call=True
    )
    def dropdown_actions(action:str, active_tab:str,):
        r"""
        documentation here
        """
        if action:

            machine_name = active_tab.split("-")[-1]
            message = f'are you sure you want to {action} {machine_name} state machine?'
            dash.set_props("modal-question-body-action-machine", {"children": message})
            dash.set_props("modal-question-action-machine", {'is_open': True})

    @app.callback(
        dash.Output("modal-error-subscription", "is_open"),
        dash.Input("close-modal-error-button-subscription", "n_clicks"),
        [dash.State("modal-error-subscription", "is_open")],
    )
    def close_error_button(n, is_open):
        r"""
        Documentation here
        """
        if n:

            return not is_open
        
        return is_open
    
    @app.callback(
        dash.Output("modal-question-action-machine", "is_open"), 
        dash.Output("question-action-machine-yes", "n_clicks"),
        dash.Output("question-action-machine-no", "n_clicks"),
        dash.Input("question-action-machine-yes", "n_clicks"), 
        dash.Input("question-action-machine-no", "n_clicks"),
        dash.Input('machine_actions_input', 'value'),
        dash.State('machines_detailed_tabs', 'active_tab'),
        dash.State("modal-question-action-machine", "is_open"),
        prevent_initial_call=True
    )
    def question_action_machine_yes_or_no(yes_n, no_n, action:str, active_tab:str, is_open):
        r"""
        Documentation here
        """
        if action:

            if yes_n:

                machine_name = active_tab.split("-")[-1]
                machine = app.automation.get_machine(name=StringType(machine_name))

                if action=="confirm_restart":
                    machine.transition(to='wait')
                elif action=="confirm_reset":
                    machine.transition(to='start')
                elif action=="deny_restart":
                    machine.transition(to=machine.last_state)
                elif action=="deny_reset":
                    machine.transition(to=machine.last_state)
                else:
                    machine.transition(to=action)

                return not is_open, 0, 0

            elif no_n:

                return not is_open, 0, 0
            
            else:

                return True, 0, 0
        
        else:

            return False, 0, 0
        
    @app.callback(
        dash.Output('machine_threshold_input', "n_submit"),
        dash.Input('machine_threshold_input', "n_submit"),
        dash.State('machine_threshold_input', 'value'),
        dash.State('machines_detailed_tabs', 'active_tab'),
        prevent_initial_call=True
    )
    def update_threshold(n, threshold:float, active_tab:str):
        r"""
        Documentation here
        """
        try:
            threshold = float(threshold)
        except Exception as err:
            return 0
        machine_name = active_tab.split("-")[-1]
        machine = app.automation.get_machine(name=StringType(machine_name))
        machine_threshold = machine.threshold.value
        if hasattr(machine_threshold, "value"):

            machine_threshold = machine_threshold.value

        if threshold != machine_threshold:

            message = f'Do you want to change machine {machine_name} threshold from [{machine_threshold}] to: {threshold}?'
            dash.set_props("modal-update-body-attr-machine", {"children": message})
            dash.set_props("modal-update-attr-machine", {'is_open': True})

        return 0

    @app.callback(
        dash.Output('machine_interval_input', "n_submit"),
        dash.Input('machine_interval_input', "n_submit"),
        dash.State('machine_interval_input', 'value'),
        dash.State('machines_detailed_tabs', 'active_tab'),
        prevent_initial_call=True
    )
    def update_machine_interval(n, interval:float, active_tab:str):
        r"""
        Documentation here
        """
        try:
            interval = float(interval)
        except Exception as err:
            return 0
        machine_name = active_tab.split("-")[-1]
        machine = app.automation.get_machine(name=StringType(machine_name))

        if interval != machine.get_interval():

            message = f'Do you want to change machine {machine_name} interval from [{machine.get_interval()}] to: {interval}?'
            dash.set_props("modal-update-body-attr-machine", {"children": message})
            dash.set_props("modal-update-attr-machine", {'is_open': True})

        return 0

    @app.callback(
        dash.Output('buffer_size_input', "n_submit"),
        dash.Input('buffer_size_input', "n_submit"),
        dash.State('buffer_size_input', 'value'),
        dash.State('machines_detailed_tabs', 'active_tab'),
        prevent_initial_call=True
    )
    def update_buffer_size(n, buffer_size:int, active_tab:str):
        r"""
        Documentation here
        """
        try:
            buffer_size = int(buffer_size)
        except Exception as err:
            return 0
        machine_name = active_tab.split("-")[-1]
        machine = app.automation.get_machine(name=StringType(machine_name))

        if buffer_size != machine.buffer_size.value:

            message = f'Do you want to change machine {machine_name} buffer size from [{machine.buffer_size.value}] to: {buffer_size}?'
            dash.set_props("modal-update-body-attr-machine", {"children": message})
            dash.set_props("modal-update-attr-machine", {'is_open': True})

        return 0

    @app.callback(
        dash.Output('on_delay_input', "n_submit"),
        dash.Input('on_delay_input', "n_submit"),
        dash.State('on_delay_input', 'value'),
        dash.State('machines_detailed_tabs', 'active_tab'),
        prevent_initial_call=True
    )
    def update_on_delay(n, on_delay:int, active_tab:str):
        r"""
        Documentation here
        """
        try:
            on_delay = int(on_delay)
        except Exception as err:
            return 0
        machine_name = active_tab.split("-")[-1]
        machine = app.automation.get_machine(name=StringType(machine_name))

        if on_delay != machine.on_delay.value:

            message = f'Do you want to change machine {machine_name} on delay from [{machine.on_delay.value}] to: {on_delay}?'
            dash.set_props("modal-update-body-attr-machine", {"children": message})
            dash.set_props("modal-update-attr-machine", {'is_open': True})

        return 0
    
    @app.callback(
        dash.Output("modal-update-attr-machine", "is_open"), 
        dash.Output("update-attr-machine-yes", "n_clicks"),
        dash.Output("update-attr-machine-no", "n_clicks"),
        dash.Output('machine_threshold_input', "value"),
        dash.Output('machine_interval_input', "value"),
        dash.Output('buffer_size_input', "value"),
        dash.Output('on_delay_input', "value"),
        dash.Input("update-attr-machine-yes", "n_clicks"), 
        dash.Input("update-attr-machine-no", "n_clicks"),
        dash.State('machines_detailed_tabs', 'active_tab'),
        dash.State("modal-update-attr-machine", "is_open"),
        dash.State('machine_threshold_input', 'value'),
        dash.State('machine_interval_input', 'value'),
        dash.State('buffer_size_input', 'value'),
        dash.State('on_delay_input', 'value'),
        prevent_initial_call=True
    )
    def update_attr_machine_yes_or_no(
        yes_n, 
        no_n, 
        active_tab:str, 
        is_open, 
        threshold:float=None, 
        interval:int=None, 
        buffer_size:int=None, 
        on_delay:int=None
    ):
        r"""
        Documentation here
        """
        machine_name = active_tab.split("-")[-1]
        machine = app.automation.get_machine(name=StringType(machine_name))
        if yes_n:
            
            if threshold:
                
                if "leak detection" in machine.classification.value.lower():

                    if machine_name.lower()=="npw":

                        if threshold > 100:

                            threshold = 100

                        elif threshold < 0:

                            threshold = 0

                        machine.wavelet.threshold_iqr = threshold

                machine.threshold.value.value = threshold
                # UPDATE DB
                if app.automation.is_db_connected():
                    app.automation.machines_engine.put(name=StringType(machine_name), threshold=FloatType(threshold))
                dash.set_props("machine_threshold_input", {"placeholder": f"Current threshold {machine.threshold.value.value}"})

            elif interval:

                machine.set_interval(interval=IntegerType(interval))
                # UPDATE DB
                if app.automation.is_db_connected():
                    app.automation.machines_engine.put(name=StringType(machine_name), machine_interval=IntegerType(interval))
                dash.set_props("machine_interval_input", {"placeholder": f"Current machine interval {machine.get_interval()}"})
                
            elif buffer_size:

                machine.set_buffer_size(size=buffer_size)
                machine.transition(to="restart")
                if app.automation.is_db_connected():
                    app.automation.machines_engine.put(name=StringType(machine_name), buffer_size=IntegerType(buffer_size))
                machine.transition(to="wait")
                # UPDATE DB
                dash.set_props("buffer_size_input", {"placeholder": f"Current buffer size {machine.buffer_size.value}"})

            elif on_delay:

                machine.on_delay.value = on_delay
                if app.automation.is_db_connected():
                    app.automation.machines_engine.put(name=StringType(machine_name), on_delay=IntegerType(on_delay))
                # UPDATE DB
                dash.set_props("on_delay_input", {"placeholder": f"Current on delay {machine.on_delay.value}"})
            
        return not is_open, 0, 0, "", "", "", ""


from ..utils.units import EngUnit, UnitSerializer, UnitError

class Temperature(EngUnit):
    """Creates a temperature object that can store a temperature value and 
    convert between units of temperature.

    :param value: [int|float] Engineering value\n
    :param unit: [str] Engineering unit\n
    :return: [Temperature Object]\n

    ```python
    >>> from automation.variables.temperature import Temperature
    >>> temperature = Temperature(value=25, unit="C")
    >>> temperature.value
    25
    >>> temperature.unit
    'C'
    >>> temperature.convert(to_unit="F")
    76.99999999999994
    >>> Temperature.convert_value(value=1.0, from_unit="C", to_unit="K")
    274.15
    >>> temperature.change_unit(unit="F")
    76.99999999999994
    >>> temperature.unit
    'F'
    >>> temperature.get_value()
    [76.99999999999994, 'F']
    >>> print(temperature)
    76.99999999999994 F
    >>> temperature2 = Temperature(value=3.0, unit="K")
    >>> temperature_result = temperature + temperature2
    >>> print(temperature_result)
    82.39999999999992 F
    >>> temperature_result = temperature * temperature2
    >>> print(temperature_result)
    894.4499999999999 K
    >>> temperature_result = temperature / temperature2
    >>> print(temperature_result)
    99.38333333333333 K
    >>> temperature_result = temperature ** temperature2
    >>> print(temperature_result)
    26503573.918374993 K

    ```
    """
    
    class Units(UnitSerializer):
        degKelvin = 'K'
        degCelsius = 'C'
        degRankine = 'R'
        degFarenheit = 'F'

    # conversions { } is not used to convert the Temperature() class because 
    # temperature is not converted with a scalar.  See the convert() function below.
    conversions = {
        'K' : 1.0,
    }

    def __init__(self, value, unit):

        if unit not in Temperature.Units.list():

            raise UnitError(f"{unit} value is not allowed for {self.__class__.__name__} object - you can use: {Temperature.Units.list()}")
        
        super(Temperature, self).__init__(value=value, unit=unit)

    def convert(self, to_unit):
        """
        Converts a temperature value from one unit of measure to another.
        Returns a float value of the temperature in requested units. 
        Returns None for incorrect value.

        Units of Measure
        ---------------
        Units of measure entered as a string.
        'K' - Degree Kelvin\n
        'F' - Degree Farenheit\n
        'R' - Degree Rankine\n
        'C' - Degree Celcius

        Parameters
        ---------------
        value : float
            Value of the temperature measurement.\n
        from_unit : str
            Unit of measurement to convert from.\n
        to_unit : str
            Unit of measurement to convert to.
        """
    
        temperature_kelvin = 0
        if self.unit.upper() == 'K':
            temperature_kelvin = self.value
        elif self.unit == 'R':
            temperature_kelvin = self.value * 5.0 / 9.0
        elif self.unit == 'C':
            temperature_kelvin = self.value + 273.15
        elif self.unit == 'F':
            temperature_kelvin = (self.value + 459.67) / 9.0 * 5.0
        else:
            return None
        
        # Return Value in Required Unit
        if to_unit == 'K':
            return float(temperature_kelvin)
        elif to_unit == 'R':
            return temperature_kelvin * 9.0 / 5.0
        elif to_unit == 'C':
            return temperature_kelvin - 273.15
        elif to_unit == 'F':
            return temperature_kelvin * 9.0 / 5.0 - 459.67
        else:
            return None  

    @classmethod
    def convert_value(cls, value, from_unit:str, to_unit:str):
        
        temperature_kelvin = 0
        if from_unit.upper() == 'K':
            temperature_kelvin = value
        elif from_unit.upper() == 'R':
            temperature_kelvin = value * 5.0 / 9.0
        elif from_unit.upper() == 'C':
            temperature_kelvin = value + 273.15
        elif from_unit.upper() == 'F':
            temperature_kelvin = (value + 459.67) / 9.0 * 5.0
        else:
            return None
        
        # Return Value in Required Unit
        if to_unit == 'K':
            return float(temperature_kelvin)
        elif to_unit == 'R':
            return temperature_kelvin * 9.0 / 5.0
        elif to_unit == 'C':
            return temperature_kelvin - 273.15
        elif to_unit == 'F':
            return temperature_kelvin * 9.0 / 5.0 - 459.67
        else:
            return None  

    def __add__(self, other):
        self_original_unit = self.unit

        self.change_unit('K')

        if other.unit in ('K', 'C'):
            new_value = self.value + other.value
        else:
            new_value = self.value + (other.value * 5.0 / 9.0)            

        new_unit = self.__class__(new_value, 'K')
        new_unit.change_unit(self_original_unit)
        return new_unit

    def __sub__(self, other):
            self_original_unit = self.unit

            self.change_unit('K')

            if other.unit in ('K', 'C'):
                new_value = self.value - other.value
            else:
                new_value = self.value - (other.value * 5.0 / 9.0)            

            new_unit = self.__class__(new_value, 'K')
            new_unit.change_unit(self_original_unit)
            return new_unit


# -*- coding: utf-8 -*-
"""automation/logger/logdict.py

This module implements a dictionary-based helper class to organize and validate
tags that need to be logged, grouping them by sampling period.
"""


class LogTable(dict):
    r"""
    A specialized dictionary for managing logging configurations.

    Structure:
    `{ period (float): [tag_name_1, tag_name_2, ...] }`

    It validates entries and provides helper methods to retrieve tags by group or individually.
    """

    def __init__(self):

        pass

    def validate(self, period, tag):
        r"""
        Validates a period-tag pair.

        **Parameters:**

        * **period** (int|float): Sampling period.
        * **tag** (str): Tag name.

        **Returns:**

        * **bool**: True if valid.
        """
        if not type(period) in [int, float]:
            return False
        
        if type(tag) != str:
            return False

        return True

    def get_groups(self):
        r"""
        Retrieves all defined sampling periods (groups).

        **Returns:**

        * **list**: List of periods.
        """
        return list(self.keys())

    def get_tags(self, group):
        r"""
        Retrieves all tags for a specific sampling period.

        **Parameters:**

        * **group** (float): Sampling period.

        **Returns:**

        * **list**: List of tag names.
        """
        return self[group]

    def get_all_tags(self):
        r"""
        Retrieves a flat list of all tags across all groups.

        **Returns:**

        * **list**: List of tag names.
        """
        result = list()

        for group in self.get_groups():

            result += self.get_tags(group)

        return result

    def get_period(self, tag):
        r"""
        Finds the sampling period for a given tag.

        **Parameters:**

        * **tag** (str): Tag name.

        **Returns:**

        * **float**: The period if found, else None.
        """
        for key, value in self.items():

            if tag in value:
                return key

    def serialize(self):
        r"""
        Returns the dictionary representation of the LogTable.
        """
        return self



import pytz
from datetime import datetime
from math import ceil
from ..singleton import Singleton
from ..tags.cvt import CVTEngine
from ..tags import Tag
from ..buffer import Buffer
from ..models import StringType
from ..logger.datalogger import DataLoggerEngine


class SubHandler(Singleton):
    r"""
    Subscription Handler. To receive events from server for a subscription
    data_change and event methods are called directly from receiving thread.
    Do not do expensive, slow or network operation there. Create another 
    thread if you need to do such a thing
    """

    def __init__(self):
        
        self.monitored_items = dict()

    def subscribe(self, subscription, client_name, node_id):
        r"""
        Documentation here
        """
        
        if client_name not in self.monitored_items:
            
            monitored_item = subscription.subscribe_data_change(
                node_id
            )

            self.monitored_items[client_name] = {
                node_id: {
                    "subscription": subscription,
                    "monitored_item": monitored_item,
                    "server": client_name
                }
            }

        else:

            if node_id not in self.monitored_items[client_name]:
            
                monitored_item = subscription.subscribe_data_change(
                    node_id
                )

                self.monitored_items[client_name].update({
                    node_id: {
                        "subscription": subscription,
                        "monitored_item": monitored_item,
                        "server": client_name
                    }
                })

    def unsubscribe_all(self):
        r"""
        Documentation here
        """

        for _, monitored_items in self.monitored_items.items():

            for _, monitored_item in monitored_items.items():
                
                item = monitored_item["monitored_item"]
                subscription = monitored_item["subscription"]
                subscription.unsubscribe(item)
                
        self.monitored_items = dict()  

    def resubscribe_all(self, client): 
        for client_name, monitored_items in self.monitored_items.items(): 
            for node_id, monitored_item in monitored_items.items(): 
                subscription = monitored_item["subscription"] 
                monitored_item["monitored_item"] = subscription.subscribe_data_change(client.get_node(node_id))   

    def datachange_notification(self, node, val, data):
        r"""
        Documentation here
        """
        pass

class SubHandlerServer(Singleton):

    def __init__(self):
        from ..core import PyAutomation
        self.app = PyAutomation()
        self.subscriptions = dict()

    def is_property(self, node):
        from ..state_machine import ua
        # A property is a variable that is part of another node (often an Object or another Variable)
        parent = node.get_parent()
        references = node.get_references(refs=ua.ObjectIds.HasProperty)
        return bool(references)

    def is_variable(self, node):
        from ..state_machine import ua
        return node.get_node_class() == ua.NodeClass.Variable

    def datachange_notification(self, node, val, data):
        from .. import SEGMENT, MANUFACTURER, PyAutomation

        app = PyAutomation()

        timestamp = data.monitored_item.Value.SourceTimestamp
        if not timestamp:
            timestamp = datetime.now(pytz.utc)
        timestamp = timestamp.replace(tzinfo=pytz.UTC)
        tag_name = node.get_display_name().Text
        
        tag = self.app.get_tag_by_name(name=tag_name)
        if tag:
            if tag.get_value()!=val:

                val = tag.value.convert_value(value=val, from_unit=tag.get_unit(), to_unit=tag.get_display_unit())
                if tag.manufacturer==MANUFACTURER and tag.segment==SEGMENT:      
                    self.app.cvt.set_value(id=tag.id, value=val, timestamp=timestamp)
                elif not MANUFACTURER and not SEGMENT:
                    self.app.cvt.set_value(id=tag.id, value=val, timestamp=timestamp) 
        else:
            
            parent = node.get_parent()
            if parent:
                machine_name = parent.get_display_name().Text
                machine = app.get_machine(name=StringType(machine_name))
                attr = getattr(machine, tag_name)
                attr.value = val
                

class DAS(Singleton):
    r"""
    Subscription Handler. To receive events from server for a subscription
    data_change and event methods are called directly from receiving thread.
    Do not do expensive, slow or network operation there. Create another 
    thread if you need to do such a thing
    """

    def __init__(self):
  
        self.monitored_items = dict()
        self.cvt = CVTEngine()
        self.logger = DataLoggerEngine()
        self.buffer = dict()

    def restart_buffer(self, tag:Tag):
        r"""
        Documentation here
        """
        scan_time = tag.get_scan_time()
        if scan_time:
            
            self.buffer[tag.get_name()].update({
                "timestamp": Buffer(size=ceil(600/ ceil(scan_time / 1000))),
                "values": Buffer(size=ceil(600 / ceil(scan_time / 1000)))
            })
        else:
            self.buffer[tag.get_name()].update({
                "timestamp": Buffer(size=600),
                "values": Buffer(size=600)
            })

    def subscribe(self, subscription, client_name, node_id):
        r"""
        Documentation here
        """
        if client_name not in self.monitored_items:
            
            monitored_item = subscription.subscribe_data_change(
                node_id
            )
            self.monitored_items[client_name] = {
                node_id.get_display_name().Text: {
                    "subscription": subscription,
                    "monitored_item": monitored_item,
                    "server": client_name,
                    "namespace": node_id.nodeid.to_string()
                }
            }

        else:

            if node_id not in self.monitored_items[client_name]:
                
                monitored_item = subscription.subscribe_data_change(
                    node_id
                )

                self.monitored_items[client_name].update({
                    node_id.get_display_name().Text: {
                        "subscription": subscription,
                        "monitored_item": monitored_item,
                        "server": client_name,
                        "namespace": node_id.nodeid.to_string()
                    }
                })
        
        ## Trying to get the value of the tag into OPCUA Client
        try:
            val = node_id.get_value()
            self.update_tag_value(node=node_id, val=val)
        except Exception:
            pass

    def unsubscribe(self, client_name:str, node_id):
        r"""
        Documentation here
        """
        if client_name in self.monitored_items:
            
            if node_id.get_display_name().Text in self.monitored_items[client_name]:
                
                node = self.monitored_items[client_name].pop(node_id.get_display_name().Text)
                item = node["monitored_item"]
                subscription = node["subscription"]
                subscription.unsubscribe(item)

    def resubscribe_all(self, client): 
        for client_name, monitored_items in self.monitored_items.items(): 
            for node_id, monitored_item in monitored_items.items(): 
                subscription = monitored_item["subscription"] 
                monitored_item["monitored_item"] = subscription.subscribe_data_change(client.get_node(node_id))

    def update_tag_value(self, node, val, timestamp=None):
        r"""
        Update tag value in CVT and buffer
        """
        from .. import SEGMENT, MANUFACTURER, TIMEZONE
        
        if not timestamp:
            timestamp = datetime.now(pytz.utc)
        timestamp = timestamp.replace(tzinfo=pytz.UTC)
        
        namespace = node.nodeid.to_string()
        tag = self.cvt.get_tag_by_node_namespace(node_namespace=namespace)
        
        if tag:
            tag_name = tag.get_name()
            val = tag.value.convert_value(value=val, from_unit=tag.get_unit(), to_unit=tag.get_display_unit())
            if tag.manufacturer==MANUFACTURER and tag.segment==SEGMENT:      
                val = self.cvt.set_value(id=tag.id, value=val, timestamp=timestamp)
            elif not MANUFACTURER and not SEGMENT:
                val = self.cvt.set_value(id=tag.id, value=val, timestamp=timestamp)
            timestamp = timestamp.astimezone(TIMEZONE)
            if tag_name in self.buffer:
                self.buffer[tag_name]["timestamp"](timestamp)
                self.buffer[tag_name]["values"](val)

    def datachange_notification(self, node, val, data):
        r"""
        Documentation here
        """
        timestamp = data.monitored_item.Value.SourceTimestamp
        self.update_tag_value(node, val, timestamp)      
        
        


from flask_restx import reqparse

login_parser = reqparse.RequestParser(bundle_errors=True)
login_parser.add_argument("username", type=str, required=False, help='Username')
login_parser.add_argument("email", type=str, required=False, help='User email')
login_parser.add_argument("password", type=str, required=True, help='User password')

signup_parser = reqparse.RequestParser(bundle_errors=True)
signup_parser.add_argument("username", type=str, required=True, help='Username')
signup_parser.add_argument("email", type=str, required=True, help="Email address")
signup_parser.add_argument("password", type=str, required=True, help="Password")
signup_parser.add_argument("name", type=str, required=False, help="User's name")
signup_parser.add_argument("lastname", type=str, required=False, help="User's lastname")

change_password_parser = reqparse.RequestParser(bundle_errors=True)
change_password_parser.add_argument("target_username", type=str, required=True, help='Username whose password will be changed')
change_password_parser.add_argument("new_password", type=str, required=True, help='New password')
change_password_parser.add_argument("current_password", type=str, required=False, help='Current password (required when changing own password)')

reset_password_parser = reqparse.RequestParser(bundle_errors=True)
reset_password_parser.add_argument("target_username", type=str, required=True, help='Username whose password will be reset')
reset_password_parser.add_argument("new_password", type=str, required=True, help='New password')

create_tpt_parser = reqparse.RequestParser(bundle_errors=True)
create_tpt_parser.add_argument("role_name", type=str, required=True, help='Role name to embed in the JWT token')


from flask import Blueprint, request
from flask_restx import abort
from flask_restx import Api as API
from ..singleton import Singleton
from functools import wraps
import logging, jwt
from ..utils.decorators import decorator
from ..dbmodels.users import Users
from ..modules.users.users import Users as CVTUsers


authorizations = {
    'apikey' : {
        'type' : 'apiKey',
        'in' : 'header',
        'name' : 'X-API-KEY'
    }
}


blueprint = Blueprint('api', __name__, url_prefix='/api')

api = API(blueprint, version='1.0', 
        title='PyAutomation API',
        description="""
        This API groups all namespaces defined in every module's resources for PyAutomation App.
        """, 
        doc='/docs',
        authorizations=authorizations
    )

users = CVTUsers()

class Api(Singleton):

    def __init__(self):

        self.app = None

    def init_app(self, app):
        r"""
        Documentation here
        """
        self.app = self.create_api(app)

        return app

    def create_api(self, app):
        r"""
        Documentation here
        """
        app.register_blueprint(blueprint)

        return api
    
    @staticmethod
    def verify_tpt(tpt:str):
        r"""
        Verify Third Party Token
        """
        from .. import server
        try:

            jwt.decode(tpt, server.config["AUTOMATION_APP_SECRET_KEY"], algorithms=["HS256"])

            return True

        except:

            return 
        
    @classmethod
    def validate_reqparser(cls, reqparser):
        def _validate_reqparser(f):
                
            @wraps(f)
            def decorated(*args, **kwargs):
                
                reqparser.parse_args() 
                result = f(*args, **kwargs)
                return result

            return decorated
        
        return _validate_reqparser
    
    @classmethod
    def token_required(cls, auth:bool=False):
        
        def _token_required(f):
            
            @wraps(f)
            def decorated(*args, **kwargs):
                try:

                    if auth:

                        token = None

                        if 'X-API-KEY' in request.headers:
                            
                            token = request.headers['X-API-KEY']

                        elif 'Authorization' in request.headers:
                            
                            token = request.headers['Authorization'].split('Token ')[-1]

                        if not token:
                            
                            return {'message' : 'Key is missing.'}, 401
                        
                        user = Users.get_or_none(token=token)

                        if user:

                            return f(*args, **kwargs)

                        if Api.verify_tpt(tpt=token):
                    
                            return f(*args, **kwargs)

                        return {'message' : 'Invalid token'}, 401                  
                
                except Exception as err:
                    logger = logging.getLogger("pyautomation")
                    logger.error(str(err))

            return decorated

        return _token_required
    
    @classmethod
    def auth_roles(cls, role_names:list[str]):
        r"""
        Decorator that restricts access to endpoints based on a list of role names.
        
        **Parameters:**
        
        * **role_names** (list[str]): List of role names allowed to access the endpoint.
        
        **Usage:**
        
        ```python
        @Api.token_required(auth=True)
        @Api.auth_roles(['admin', 'supervisor'])
        def post(self):
            # Only users with role 'admin' or 'supervisor' can access
            pass
        ```
        """
        def _auth_roles(f):
            @wraps(f)
            def decorated(*args, **kwargs):
                try:
                    token = None
                    
                    if 'X-API-KEY' in request.headers:
                        token = request.headers['X-API-KEY']
                    elif 'Authorization' in request.headers:
                        token = request.headers['Authorization'].split('Token ')[-1]
                    
                    if not token:
                        return {'message': 'Token is required'}, 401
                    
                    # Get user from token
                    current_user = users.get_active_user(token=token)
                    
                    if not current_user:
                        # Try database user
                        db_user = Users.get_or_none(token=token)
                        if db_user:
                            # Get role from database
                            role_name = db_user.role.name.upper()
                            if role_name in [r.upper() for r in role_names]:
                                return f(*args, **kwargs)
                        return {'message': 'Invalid token or insufficient permissions'}, 401
                    
                    # Check if user's role is in the allowed list
                    user_role_name = current_user.role.name.upper()
                    allowed_roles = [r.upper() for r in role_names]
                    
                    if user_role_name in allowed_roles:
                        return f(*args, **kwargs)
                    
                    return {'message': f'Access denied. Required roles: {role_names}'}, 403
                    
                except Exception as err:
                    logger = logging.getLogger("pyautomation")
                    logger.error(str(err))
                    return {'message': 'Internal server error'}, 500
            
            return decorated
        return _auth_roles
    
    @classmethod
    def auth_role_level(cls, max_level:int):
        r"""
        Decorator that restricts access to endpoints based on role level.
        Users with role_level <= max_level are allowed access.
        
        **Parameters:**
        
        * **max_level** (int): Maximum role level allowed (inclusive). Lower numbers = higher privilege.
        
        **Usage:**
        
        ```python
        @Api.token_required(auth=True)
        @Api.auth_role_level(1)  # Only admin (level 1) and sudo (level 0) can access
        def post(self):
            # Only users with role_level <= 1 can access
            pass
        ```
        """
        def _auth_role_level(f):
            @wraps(f)
            def decorated(*args, **kwargs):
                try:
                    token = None
                    
                    if 'X-API-KEY' in request.headers:
                        token = request.headers['X-API-KEY']
                    elif 'Authorization' in request.headers:
                        token = request.headers['Authorization'].split('Token ')[-1]
                    
                    if not token:
                        return {'message': 'Token is required'}, 401
                    
                    # Get user from token
                    current_user = users.get_active_user(token=token)
                    
                    if not current_user:
                        # Try database user
                        db_user = Users.get_or_none(token=token)
                        if db_user:
                            # Get role level from database
                            role_level = db_user.role.level
                            if role_level <= max_level:
                                return f(*args, **kwargs)
                        return {'message': 'Invalid token or insufficient permissions'}, 401
                    
                    # Check if user's role level is <= max_level
                    user_role_level = current_user.role.level
                    
                    if user_role_level <= max_level:
                        return f(*args, **kwargs)
                    
                    return {'message': f'Access denied. Required role level: <= {max_level}'}, 403
                    
                except Exception as err:
                    logger = logging.getLogger("pyautomation")
                    logger.error(str(err))
                    return {'message': 'Internal server error'}, 500
            
            return decorated
        return _auth_role_level
    
    @classmethod
    def get_current_user(cls):

        token = None

        if 'X-API-KEY' in request.headers:
                            
            token = request.headers['X-API-KEY']

        elif 'Authorization' in request.headers:
            
            token = request.headers['Authorization'].split('Token ')[-1]

        if token:

            return users.get_active_user(token=token)

        return None


from flask_restx import Namespace, Resource, fields, reqparse
from .... import PyAutomation
from automation.alarms import AlarmState
from ....extensions.api import api
from ....extensions import _api as Api


ns = Namespace('Alarms', description='Alarm Management Resources')
app = PyAutomation()

# Models
shelve_alarm_resource_by_name_model = api.model("shelve_alarm_resource_by_name_model",{
    'seconds': fields.Integer(required=False, description='Shelve duration in seconds'),
    'minutes': fields.Integer(required=False, description='Shelve duration in minutes'),
    'hours': fields.Integer(required=False, description='Shelve duration in hours'),
    'days': fields.Integer(required=False, description='Shelve duration in days'),
    'weeks': fields.Integer(required=False, description='Shelve duration in weeks')
})

append_alarm_resource_model = api.model("append_alarm_resource_model",{
    'name': fields.String(required=True, description='Unique Alarm Name'),
    'tag': fields.String(required=True, description='Tag name to associate with the alarm'),
    'description': fields.String(required=False, description='Alarm description'),
    'type': fields.String(required=True, description='Alarm Type. Allowed values: ["HIGH-HIGH", "HIGH", "BOOL", "LOW", "LOW-LOW"]'),
    'trigger_value': fields.Float(required=True, description="Value that triggers the alarm")
})

# Parsers
shelve_alarm_parser = reqparse.RequestParser()
shelve_alarm_parser.add_argument("seconds", type=int, required=False, help='Shelve time in seconds', default=0)
shelve_alarm_parser.add_argument("minutes", type=int, required=False, help='Shelve time in minutes', default=0)
shelve_alarm_parser.add_argument("hours", type=int, required=False, help='Shelve time in hours', default=0)
shelve_alarm_parser.add_argument("days", type=int, required=False, help='Shelve time in days', default=0)
shelve_alarm_parser.add_argument("weeks", type=int, required=False, help='Shelve time in weeks', default=0)

append_alarm_parser = reqparse.RequestParser()
append_alarm_parser.add_argument('name', type=str, required=True, help='Alarm Name')
append_alarm_parser.add_argument('tag', type=str, required=True, help='Tag to whom the alarm will be subscribed')
append_alarm_parser.add_argument('description', type=str, required=False, help='Alarm description', default="")
append_alarm_parser.add_argument('type', type=str, required=True, help='Alarm Type', choices=["HIGH-HIGH", "HIGH", "BOOL", "LOW", "LOW-LOW"])
append_alarm_parser.add_argument('trigger_value', type=float, required=True, help="Alarm trigger value")


@ns.route('/')
class AlarmsCollection(Resource):

    @api.doc(security='apikey', description="Retrieves a list of all configured alarms.")
    @api.response(200, "Success")
    @Api.token_required(auth=True)
    def get(self):
        """
        Get all alarms.

        Retrieves a list of all alarms currently defined in the system.
        """
        return app.alarm_manager.serialize(), 200
    
@ns.route('/active_alarms')
class ActiveAlarmsCollection(Resource):

    @api.doc(security='apikey', description="Checks if there are any active alarms.")
    @api.response(200, "Success", model=fields.Boolean)
    @Api.token_required(auth=True)
    def get(self):
        """
        Check for active alarms.

        Returns True if there is at least one alarm in an active state, False otherwise.
        """
        if app.alarm_manager.get_lasts_active_alarms(lasts=1):
            
            return True, 200
        
        return False, 200

    
@ns.route('/<id>')
@api.param('id', 'The alarm identifier')
class AlarmResource(Resource):

    @api.doc(security='apikey', description="Retrieves an alarm by its ID.")
    @api.response(200, "Success")
    @api.response(400, "Alarm not found")
    @Api.token_required(auth=True)
    def get(self, id):
        """
        Get alarm by ID.

        Retrieves detailed information about a specific alarm using its unique identifier.
        """
        alarm = app.get_alarm(id)

        if alarm:
        
            return alarm.serialize(), 200

        return {'message': f"Alarm ID {id} does not exist"}, 400
    

@ns.route('/name/<alarm_name>')
@api.param('alarm_name', 'The alarm name')
class AlarmByNameResource(Resource):

    @api.doc(security='apikey', description="Retrieves an alarm by its name.")
    @api.response(200, "Success")
    @api.response(400, "Alarm not found")
    @Api.token_required(auth=True)
    def get(self, alarm_name):
        """
        Get alarm by name.

        Retrieves detailed information about a specific alarm using its name.
        """
        alarm = app.alarm_manager.get_alarm_by_name(alarm_name)

        if alarm:

            return alarm.serialize(), 200 
        
        return {'message': f"Alarm Name {alarm_name} does not exist"}, 400
    
    
@ns.route('/acknowledge/<alarm_name>')
@api.param('alarm_name', 'The alarm name')
class AckAlarmByNameResource(Resource):
    
    @api.doc(security='apikey', description="Acknowledges an active alarm.")
    @api.response(200, "Alarm acknowledged successfully")
    @api.response(400, "Alarm not found or not in a state to be acknowledged")
    @Api.token_required(auth=True)
    def post(self, alarm_name:str):
        """
        Acknowledge alarm.

        Acknowledges an alarm that is in an Unacknowledged or Return to Normal Unacknowledged state.
        """
        result = dict()
        alarm = app.alarm_manager.get_alarm_by_name(alarm_name)

        if alarm:

            if alarm.state in [AlarmState.UNACK, AlarmState.RTNUN]:
                user = Api.get_current_user()
                alarm.acknowledge(user=user)
                result['message'] = f"{alarm.name} was acknowledged successfully"
                result['data'] = alarm.serialize()

                return result, 200

            return {'message': f"Alarm Name {alarm_name} is not in Unacknowledged state"}, 400

        return {'message': f"Alarm Name {alarm_name} does not exist"}, 400

    
@ns.route('/acknowledge_all')
class AckAllAlarmsResource(Resource):

    @api.doc(security='apikey', description="Acknowledges all active alarms.")
    @api.response(200, "All alarms acknowledged successfully")
    @Api.token_required(auth=True)
    def post(self):
        """
        Acknowledge all alarms.

        Iterates through all alarms and acknowledges those that are in an Unacknowledged state.
        """
        alarms = app.alarm_manager.get_alarms()

        for _, alarm in alarms.items():

            if alarm.state in [AlarmState.UNACK, AlarmState.RTNUN]:
                
                user = Api.get_current_user()
                alarm.acknowledge(user=user)
        
        result = {
            'message': "Alarms were acknowledged successfully"
        }
        
        return result, 200
    

@ns.route('/designed_suppression/<alarm_name>')
@api.param('alarm_name', 'The alarm name')
class SuppressByDesignAlarmByNameResource(Resource):
    
    @api.doc(security='apikey', description="Suppresses an alarm by design.")
    @api.response(200, "Alarm suppressed successfully")
    @api.response(400, "Alarm not found")
    @Api.token_required(auth=True)
    def post(self, alarm_name:str):
        """
        Suppress alarm by design.

        Places the alarm into a 'Designed Suppression' state, preventing it from triggering.
        """
        result = dict()
        alarm = app.alarm_manager.get_alarm_by_name(alarm_name)

        if alarm:
            user = Api.get_current_user()
            alarm.designed_suppression(user=user)
            result['message'] = f"{alarm.name} was suppressed by design successfully"
            result['data'] = alarm.serialize()

            return result, 200

        return {'message': f"Alarm Name {alarm_name} does not exist"}, 400


@ns.route('/designed_unsuppression/<alarm_name>')
@api.param('alarm_name', 'The alarm name')
class DesignedUnsuppressionAlarmByNameResource(Resource):
    
    @api.doc(security='apikey', description="Unsuppresses an alarm that was suppressed by design.")
    @api.response(200, "Alarm unsuppressed successfully")
    @api.response(400, "Alarm not found or not in suppressed state")
    @Api.token_required(auth=True)
    def post(self, alarm_name:str):
        """
        Unsuppress alarm by design.

        Returns an alarm from 'Designed Suppression' state back to normal operation.
        """
        result = dict()
        alarm = app.alarm_manager.get_alarm_by_name(alarm_name)

        if alarm:

            if alarm.state in [AlarmState.DSUPR]:
                user = Api.get_current_user()
                alarm.designed_unsuppression(user=user)
                result['message'] = f"{alarm.name} was unsuppressed by design successfully"
                result['data'] = alarm.serialize()

                return result, 200

            return {'message': f"You cannot unsuppress by design an alarm if not in suppress state"}, 400

        return {'message': f"Alarm Name {alarm_name} does not exist"}, 400


@ns.route('/remove_from_service/<alarm_name>')
@api.param('alarm_name', 'The alarm name')
class OutOfServiceAlarmByNameResource(Resource):
    
    @api.doc(security='apikey', description="Removes an alarm from service.")
    @api.response(200, "Alarm removed from service successfully")
    @api.response(400, "Alarm not found")
    @Api.token_required(auth=True)
    def post(self, alarm_name:str):
        """
        Remove alarm from service.

        Places the alarm into 'Out Of Service' state, disabling it completely.
        """
        result = dict()
        alarm = app.alarm_manager.get_alarm_by_name(alarm_name)

        if alarm:
            user = Api.get_current_user()
            alarm.remove_from_service(user=user)
            result['message'] = f"{alarm.name} was pushed in out of service successfully"
            result['data'] = alarm.serialize()

            return result, 200

        return {'message': f"Alarm Name {alarm_name} does not exist"}, 400


@ns.route('/shelve/<alarm_name>')
@api.param('alarm_name', 'The alarm name')
class ShelveAlarmByNameResource(Resource):
    
    @api.doc(security='apikey', description="Shelves an alarm for a specified duration.")
    @api.response(200, "Alarm shelved successfully")
    @api.response(400, "Alarm not found")
    @Api.token_required(auth=True)
    @ns.expect(shelve_alarm_resource_by_name_model)
    def post(self, alarm_name:str):
        """
        Shelve alarm.

        Temporarily suppresses an alarm for a specified duration (seconds, minutes, hours, days, weeks).
        """
        result = dict()
        args = shelve_alarm_parser.parse_args()
        alarm = app.alarm_manager.get_alarm_by_name(alarm_name)
        if alarm:
            user = Api.get_current_user()
            alarm.shelve(
                user=user,
                **args
            )
            result['message'] = f"{alarm.name} was shelved successfully"
            result['data'] = alarm.serialize()

            return result, 200

        return {'message': f"Alarm Name {alarm_name} does not exist"}, 400


@ns.route('/return_to_service/<alarm_name>')
@api.param('alarm_name', 'The alarm name')
class ReturnToServiceAlarmByNameResource(Resource):
    
    @api.doc(security='apikey', description="Returns an alarm to service.")
    @api.response(200, "Alarm returned to service successfully")
    @api.response(400, "Alarm not found or not in Out of Service state")
    @Api.token_required(auth=True)
    def post(self, alarm_name:str):
        """
        Return alarm to service.

        Returns an alarm from 'Out Of Service' state back to normal operation.
        """
        result = dict()
        alarm = app.alarm_manager.get_alarm_by_name(alarm_name)

        if alarm:

            if alarm.state in [AlarmState.OOSRV]:
                user = Api.get_current_user()
                alarm.return_to_service(user=user)
                result['message'] = f"{alarm.name} was returned to service successfully"
                result['data'] = alarm.serialize()

                return result, 200

            return {'message': f"You cannot returned to service an alarm if not in out of service state"}, 400

        return {'message': f"Alarm Name {alarm_name} does not exist"}, 400



# -*- coding: utf-8 -*-
"""automation/workers/logger.py

This module implements the Logger Worker, responsible for persisting data to the database.
"""
import logging, time, datetime, os, shutil
from .worker import BaseWorker
from ..managers import DBManager
from ..opcua.models import Client
from ..logger.datalogger import DataLoggerEngine
from ..tags.cvt import CVTEngine
import sqlite3
from peewee import SqliteDatabase
from ..dbmodels.tags import TagValue
from ..dbmodels.alarms import AlarmSummary
from ..dbmodels.events import Events
from ..dbmodels.logs import Logs


class LoggerWorker(BaseWorker):
    r"""
    A background worker thread that handles database operations.

    It performs the following tasks:
    1. Periodically writes buffered tag data to the database.
    2. Manages SQLite database backups and maintenance (vacuuming).
    3. Handles database reconnection logic.
    4. Checks and maintains OPC UA client connections.
    """

    def __init__(self, manager:DBManager, period:float=10.0):
        r"""
        Initializes the LoggerWorker.

        **Parameters:**

        * **manager** (DBManager): The database manager instance.
        * **period** (float): The execution interval in seconds.
        """

        super(LoggerWorker, self).__init__()
        
        self._manager = manager
        self._period = period
        self.logger = DataLoggerEngine()
        self.cvt = CVTEngine()
        self.sqlite_db = None
        self.sqlite_db_name = None

    def sqlite_db_backup(self):
        r"""
        Performs a backup of the SQLite database if it exceeds 1GB.

        It copies the database file to the `db/backups` directory, clears the
        historical tables (TagValue, AlarmSummary, Events, Logs), and performs a VACUUM
        command to reclaim disk space.
        """
        if self.sqlite_db:
            file_size_mb = os.path.getsize(self.sqlite_db_name) / 1024 / 1024 
            if file_size_mb > 1 * 1024: # 1 Gb: 
                timestamp = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
                name = self.sqlite_db_name.split(".db")[0]
                name = name.split(os.path.sep)[-1]
                backup_file = os.path.join(".", "db", "backups", f"{name}_{timestamp}.db")
                shutil.copy2(os.path.join(".", "db", "app.db"), backup_file)
                logger = logging.getLogger("pyautomation")
                logger.info(f"Backup creado: {backup_file}")
                # Empty TagValue 
                query = TagValue.delete()
                query.execute()
                # Empty Alarm Summary
                query = AlarmSummary.delete()
                query.execute()
                # Empty Events
                query = Events.delete()
                query.execute()
                # Empty Logs
                query = Logs.delete()
                query.execute()
                # Execute Vacuum to compact DB
                self.sqlite_db.close()
                conn = sqlite3.connect(self.sqlite_db_name)
                cur = conn.cursor()
                cur.execute("VACUUM;")
                conn.commit()
                conn.close()
                # Reopen DB connection
                from ..dbmodels import proxy
                self._db = self._manager.get_db()
                proxy.initialize(self._db)
                
        else:
            db = self.logger.logger.get_db()
            if db:
                if isinstance(db, SqliteDatabase):
                    self.sqlite_db = db
                    self.sqlite_db_name = db.database


    def check_opcua_connection(self):
        r"""
        Checks the status of OPC UA clients and attempts reconnection if necessary.
        """
        from automation import PyAutomation
        app = PyAutomation()
        if app.opcua_client_manager._clients:
            for _, opcua_client in app.opcua_client_manager._clients.items():

                if isinstance(opcua_client, Client):
                    
                    opcua_client.reconnect()
        else:
            
            app.load_opcua_clients_from_db()

    def get_tags_from_queue(self, _queue):
        r"""
        Retrieves tag data from the queue and filters it based on configuration.

        **Parameters:**

        * **_queue** (Queue): The queue containing tag updates.

        **Returns:**

        * **list**: A list of tag data dictionaries ready for insertion.
        """
        from .. import SEGMENT, MANUFACTURER
        tags = list()
        while not _queue.empty():

            item = _queue.get(block=False)
            tag_name = item["tag"]
            tag = self.cvt.get_tag_by_name(name=tag_name)
            if tag:

                if tag.manufacturer==MANUFACTURER and tag.segment==SEGMENT:

                    value = item['value']
                    timestamp = item["timestamp"]
                    tags.append({"tag":tag_name, "value":value, "timestamp":timestamp})

                elif not MANUFACTURER and not SEGMENT:

                    value = item['value']
                    timestamp = item["timestamp"]
                    tags.append({"tag":tag_name, "value":value, "timestamp":timestamp})

        return tags

    def reconnect_to_db(self):
        r"""
        Attempts to reconnect to the database if the connection is lost.
        """
        from automation import PyAutomation
        app = PyAutomation()
        
        if self.db_reconnection:
            
            logging.critical("Trying reconnect to DB...")
        
        self.db_reconnection = False
        db_connected = app.reconnect_to_db()
                
        if db_connected:
            
            logging.critical("Reconnection successfully")
            self.db_reconnection = True

    def run(self):
        r"""
        Main worker loop.

        Continuously:
        1. Checks database connectivity.
        2. Backs up SQLite DB if needed.
        3. Writes queued tags to the database.
        4. Reconnects to DB if connection lost.
        5. Checks OPC UA connections.
        6. Sleeps for the configured period.
        """       
        _queue = self._manager.get_queue()
        self.db_reconnection = True

        while True:

            if self.db_reconnection:

                db_connection = self.logger.logger.check_connectivity()
            
                if db_connection:
                    self.sqlite_db_backup()
                    tags = self.get_tags_from_queue(_queue=_queue)
            
                    if tags:
                        
                        self.logger.write_tags(tags=tags)
                    
                else:

                    self.reconnect_to_db()
            
            else:

                self.reconnect_to_db()

            self.check_opcua_connection()

            if self.stop_event.is_set():
                logging.critical("Alarm worker shutdown successfully!")
                break

            time.sleep(self._period)



# -*- coding: utf-8 -*-
"""automation/models.py

This module implements data models used within state machines and core logic to represent
properties and process variables with strong typing and event handling.
"""
from .modules.users.users import User
from .tags.tag import Tag
from .utils.decorators import set_event, logging_error_handler
from datetime import datetime, timezone


FLOAT = "float"
INTEGER = "int"
BOOL = "bool"
STRING = "str"

class PropertyType:

    """
    Abstract base class for property types.

    Wraps a value with type information and unit. Used for state machine attributes
    that need to be serialized or monitored.
    """

    def __init__(self, _type, default=None, unit=None):
        from .tags.cvt import CVTEngine
        from .opcua.subscription import DAS
        self._type = _type
        self.unit = unit
        self.__value = default
        self.cvt = CVTEngine()
        self.das = DAS()

    @property
    def value(self):
        r"""
        Gets the stored value.
        """
        return self.__value

    @value.setter
    def value(self, value):
        
        self.__value = value
    
    @set_event(message=f"Attribute updated", classification="State Machine", priority=2, criticity=3)
    def set_value(self, value, user:User=None, name:str=None, machine=None):
        r"""
        Sets the value and triggers updates (CVT, Events, SocketIO).

        **Parameters:**

        * **value**: The new value object.
        * **user** (User, optional): User initiating the change.
        * **name** (str, optional): Attribute name.
        * **machine** (StateMachine, optional): The parent state machine instance.
        """
        
        if isinstance(self, ProcessType):

            if not self.read_only:

                if hasattr(self, "tag"):

                    if self.tag:

                        if hasattr(machine, "data_timestamp"):
                            timestamp = machine.data_timestamp
                        else:
                            timestamp = datetime.now(timezone.utc)
                        val = self.tag.value.convert_value(value=value.value, from_unit=self.tag.get_unit(), to_unit=self.tag.get_display_unit())
                        self.tag.value.set_value(value=val, unit=self.tag.get_display_unit()) 
                        self.cvt.set_value(id=self.tag.id, value=val, timestamp=timestamp)
                        if self.tag.get_name() in self.das.buffer:
                            self.das.buffer[self.tag.get_name()]["timestamp"](timestamp)
                            self.das.buffer[self.tag.get_name()]["values"](val)
        
        if self.value:
            if value.value!=self.value.value:
                if machine:
                    if machine.sio:
                        if name:
                            payload = {machine.name.value: {name: value.value}}
                            machine.sio.emit("on.machine.property", data=payload)
                            machine.sio.emit("on.machine", data=machine.serialize())

        if name=="machine_interval":
            
            return value, f"{name} To: {value.value} s."
        
        self.value = value
        
        return value, f"{name} To: {value.value}"


class StringType(PropertyType):

    """
    Represents a String property.
    """

    def __init__(self, default=None, unit=None):

        super(StringType, self).__init__(STRING, default, unit)


class FloatType(PropertyType):

    """
    Represents a Float property.
    """

    def __init__(self, default=None, unit=None):

        super(FloatType, self).__init__(FLOAT, default, unit)


class IntegerType(PropertyType):

    """
    Represents an Integer property.
    """

    def __init__(self, default=None, unit=None):

        super(IntegerType, self).__init__(INTEGER, default, unit)

        
class BooleanType(PropertyType):

    """
    Represents a Boolean property.
    """

    def __init__(self, default=None, unit=None):

        super(BooleanType, self).__init__(BOOL, default, unit)


class ProcessType(FloatType):

    """
    Represents a process variable linked to a Tag.

    **Attributes:**

    * **read_only** (bool): If True, value comes from CVT (input). If False, writes to CVT (output).
    * **tag** (Tag): The associated Tag object.
    """

    def __init__(self, tag:Tag|None=None, default=None, read_only:bool=True, unit:str=None):
        
        self.tag = tag
        self.read_only = read_only
        
        super(ProcessType, self).__init__(default=default, unit=unit)
        
    @logging_error_handler
    def serialize(self):
        r"""
        Serializes the process variable.

        **Returns:**

        * **dict**: {value, unit, tag, read_only}
        """
        tag = None
        if self.tag:

            tag = self.tag.serialize()

        value = None
        if self.value:
            
            if isinstance(self.value, (bool, float, int, str)):

                value = self.value

            elif isinstance(self.value, (BooleanType, FloatType, IntegerType, StringType)):

                value = self.value.value

        return {
            "value": value,
            "unit": self.unit,
            "tag": tag,
            "read_only": self.read_only
        }



from peewee import Proxy, Model

proxy = Proxy()

SQLITE = 'sqlite'
MYSQL = 'mysql'
POSTGRESQL = 'postgresql'


class BaseModel(Model):
    r"""
    Base model class for all database models.

    It provides common utility methods for CRUD operations and serialization.
    Inherits from `peewee.Model` and uses a dynamic database proxy.
    """

    @classmethod
    def read(cls, id:int|str):
        r"""
        Select a single record by its ID.

        **Parameters:**

        * **id** (int|str): The primary key of the record.

        **Returns:**

        * **Model**: The model instance if found, otherwise None.
        """
        query = cls.select().where(cls.id == id).get_or_none()

        if query:
            
            return query

    @classmethod
    def read_all(cls):
        r"""
        Retrieves all records from the table.

        **Returns:**

        * **list**: A list of serialized dictionaries representing all records.
        """
        data = [query.serialize() for query in cls.select()]

        return data

    @classmethod
    def put(cls, id:str, **fields):
        r"""
        Updates a record by its ID.

        **Parameters:**

        * **id** (str): The primary key of the record to update.
        * **fields**: Keyword arguments representing the fields to update.

        **Returns:**

        * **query**: The execution result.
        """     
        if cls.id_exists(id):
            
            query = cls.update(**fields).where(cls.id == id)
            query.execute()
            return query            

    @classmethod
    def id_exists(cls, id:str|int)->bool|None:
        r"""
        Checks if a record exists by its ID.

        **Parameters:**

        * **id** (int|str): Record ID.

        **Returns:**

        * **bool**: True if the record exists, False otherwise.
        """
        return True if cls.get_or_none(id=id) else False

    class Meta:
        database = proxy


import dash
from .tags import init_callback as init_callback_tags
from .opcua import init_callback as init_callback_opcua
from .alarms import init_callback as init_callback_alarms
from .alarms_summary import init_callback as init_callback_alarms_summary
from .trends import init_callback as init_callback_trends
from .db import init_callback as init_callback_db
from .machines import init_callback as init_callback_machines
from .machines_detailed import init_callback as init_callback_machines_detailed
from .filter import init_callback as init_callback_filter
from .opcua_server import init_callback as init_callback_opcua_server


def init_callbacks(app:dash.Dash):
    r"""
    Documentation here
    """

    init_callback_tags(app=app)
    init_callback_opcua(app=app)
    init_callback_alarms(app=app)
    init_callback_alarms_summary(app=app)
    init_callback_trends(app=app)
    init_callback_db(app=app)
    init_callback_machines(app=app)
    init_callback_machines_detailed(app=app)
    init_callback_filter(app=app)
    init_callback_opcua_server(app=app)